package Bottom;

import static Bottom.Domain.*;
import Common.ArrowType;
import static Common.ArrowType.*;
import Common.ParseArrowType;
import Common.IArrow;


public enum Arrow implements Common.IArrow<Arrow,Domain> {

    //  Tuple Int
    Int000         (Void, Int, variable, 1, false, "Int",
                                         "$Int.select(t->t.toTable().exists()).exists()"),
    toTable002     (Int, $Int, noargs, 2, false, "toTable()",
                                         "$Int.select(t->t.toTable().exists()).exists()"),
    id003          (Int, Int, noargs, 3, false, "id()",
                                         "$Int.select(t->t.id().toTable().exists()).exists()"),
    eq004          (Int, Bool, hasArgs, Int, 4, false, "eq(",
                                         "5.eq(5)"),
    geq005         (Int, Bool, hasArgs, Int, 5, false, "geq(",
                                         "5.geq(5)"),
    leq006         (Int, Bool, hasArgs, Int, 6, false, "leq(",
                                         "5.leq(5)"),
    gtr007         (Int, Bool, hasArgs, Int, 7, false, "gtr(",
                                         "5.gtr(5)"),
    lss008         (Int, Bool, hasArgs, Int, 8, false, "lss(",
                                         "5.lss(5)"),

    //  Table $Int
    $Int009        (Void, $Int, variable, 9, false, "$Int",
                                         "$Int.exists()"),
    exists010      ($Int, Bool, noargs, 10, false, "exists()",
                                         "$Int.exists()"),
    count011       ($Int, Int, noargs, 11, false, "count()",
                                         "$Int.count().eq(5)"),
    id012          ($Int, $Int, noargs, 12, false, "id()",
                                         "$Int.id().exists()"),
    select013      ($Int, $Int, hasArgs, Bool, Int, 13, false, "select(",
                                         "$Int.select(t->t.toTable().exists()).exists()"),

    //  Tuple Str
    Str000         (Void, Str, variable, 14, false, "Str",
                                         "$Str.select(t->t.toTable().exists()).exists()"),
    toTable015     (Str, $Str, noargs, 15, false, "toTable()",
                                         "$Str.select(t->t.toTable().exists()).exists()"),
    id016          (Str, Str, noargs, 16, false, "id()",
                                         "$Str.select(t->t.id().toTable().exists()).exists()"),
    equals017      (Str, Bool, hasArgs, Str, 17, false, "equals(",
                                         "\"d\".equals(\"d\")"),

    //  Table $Str
    $Str018        (Void, $Str, variable, 18, false, "$Str",
                                         "$Str.exists()"),
    exists019      ($Str, Bool, noargs, 19, false, "exists()",
                                         "$Str.exists()"),
    count020       ($Str, Int, noargs, 20, false, "count()",
                                         "$Str.count().eq(5)"),
    id021          ($Str, $Str, noargs, 21, false, "id()",
                                         "$Str.id().exists()"),
    select022      ($Str, $Str, hasArgs, Bool, Str, 22, false, "select(",
                                         "$Str.select(t->t.toTable().exists()).exists()"),

    //  Tuple Bool
    Bool000        (Void, Bool, variable, 23, false, "Bool",
                                         "$Bool.select(t->t.toTable().exists()).exists()"),
    toTable024     (Bool, $Bool, noargs, 24, false, "toTable()",
                                         "$Bool.select(t->t.toTable().exists()).exists()"),
    id025          (Bool, Bool, noargs, 25, false, "id()",
                                         "$Bool.select(t->t.id().toTable().exists()).exists()"),
    and026         (Bool, Bool, hasArgs, Bool, 26, false, "and(",
                                         "5.eq(5).and(4.eq(4))"),
    or027          (Bool, Bool, hasArgs, Bool, 27, false, "or(",
                                         "5.eq(5).or(4.eq(4))"),
    not028         (Bool, Bool, hasArgs, Bool, 28, false, "not(",
                                         "5.eq(5).not(4.eq(4))"),

    //  Table $Bool
    $Bool029       (Void, $Bool, variable, 29, false, "$Bool",
                                         "$Bool.exists()"),
    exists030      ($Bool, Bool, noargs, 30, false, "exists()",
                                         "$Bool.exists()"),
    count031       ($Bool, Int, noargs, 31, false, "count()",
                                         "$Bool.count().eq(5)"),
    id032          ($Bool, $Bool, noargs, 32, false, "id()",
                                         "$Bool.id().exists()"),
    select033      ($Bool, $Bool, hasArgs, Bool, Bool, 33, false, "select(",
                                         "$Bool.select(t->t.toTable().exists()).exists()"),

    //  Tuple Pair
    Pair000        (Void, Pair, variable, 34, false, "Pair",
                                         "$Pair.select(t->t.toTable().exists()).exists()"),
    toTable035     (Pair, $Pair, noargs, 35, false, "toTable()",
                                         "$Pair.select(t->t.toTable().exists()).exists()"),
    id036          (Pair, Pair, noargs, 36, false, "id()",
                                         "$Pair.select(t->t.id().toTable().exists()).exists()"),
    first037       (Pair, Str, field, 37, false, "first",
                                         "$Pair.select(t->t.first.equals(\"d\")).exists()"),
    last038        (Pair, Str, field, 38, false, "last",
                                         "$Pair.select(t->t.last.equals(\"d\")).exists()"),

    //  Table $Pair
    $Pair039       (Void, $Pair, variable, 39, false, "$Pair",
                                         "$Pair.exists()"),
    exists040      ($Pair, Bool, noargs, 40, false, "exists()",
                                         "$Pair.exists()"),
    count041       ($Pair, Int, noargs, 41, false, "count()",
                                         "$Pair.count().eq(5)"),
    id042          ($Pair, $Pair, noargs, 42, false, "id()",
                                         "$Pair.id().exists()"),
    select043      ($Pair, $Pair, hasArgs, Bool, Pair, 43, false, "select(",
                                         "$Pair.select(t->t.toTable().exists()).exists()"),
    first044       ($Pair, $Str, noargs, 44, false, "first()",
                                         "$Pair.first().exists()"),
    last045        ($Pair, $Str, noargs, 45, false, "last()",
                                         "$Pair.last().exists()"),

    //  Tuple Bank
    Bank046        (Void, Bank, variable, 46, false, "Bank",
                                         "$Bank.select(t->t.toTable().exists()).exists()"),
    toTable047     (Bank, $Bank, noargs, 47, false, "toTable()",
                                         "$Bank.select(t->t.toTable().exists()).exists()"),
    id048          (Bank, Bank, noargs, 48, false, "id()",
                                         "$Bank.select(t->t.id().toTable().exists()).exists()"),
    name049        (Bank, Str, field, 49, true, "name",
                                         "$Bank.select(t->t.name.equals(\"d\")).exists()"),
    vb_050         (Bank, $Vector, noargs, 50, true, "vb_()",
                                         "$Bank.select(t->t.vb_().exists()).exists()"),
    link2v051      (Bank, Vector, noargs, 51, false, "link2v()",
                                         "$Bank.select(t->t.link2v().toTable().exists()).exists()"),

    //  Table $Bank
    $Bank052       (Void, $Bank, variable, 52, false, "$Bank",
                                         "$Bank.exists()"),
    exists053      ($Bank, Bool, noargs, 53, false, "exists()",
                                         "$Bank.exists()"),
    count054       ($Bank, Int, noargs, 54, false, "count()",
                                         "$Bank.count().eq(5)"),
    id055          ($Bank, $Bank, noargs, 55, false, "id()",
                                         "$Bank.id().exists()"),
    select056      ($Bank, $Bank, hasArgs, Bool, Bank, 56, false, "select(",
                                         "$Bank.select(t->t.toTable().exists()).exists()"),
    name057        ($Bank, $Str, noargs, 57, true, "name()",
                                         "$Bank.name().exists()"),
    vb_058         ($Bank, $Vector, noargs, 58, true, "vb_()",
                                         "$Bank.vb_().exists()"),

    //  Tuple Account
    Account059     (Void, Account, variable, 59, false, "Account",
                                         "$Account.select(t->t.toTable().exists()).exists()"),
    toTable060     (Account, $Account, noargs, 60, false, "toTable()",
                                         "$Account.select(t->t.toTable().exists()).exists()"),
    id061          (Account, Account, noargs, 61, false, "id()",
                                         "$Account.select(t->t.id().toTable().exists()).exists()"),
    accNum062      (Account, Int, field, 62, true, "accNum",
                                         "$Account.select(t->t.accNum.eq(5)).exists()"),
    av063          (Account, $Vector, noargs, 63, true, "av()",
                                         "$Account.select(t->t.av().exists()).exists()"),
    link2v064      (Account, Vector, noargs, 64, false, "link2v()",
                                         "$Account.select(t->t.link2v().toTable().exists()).exists()"),

    //  Table $Account
    $Account065    (Void, $Account, variable, 65, false, "$Account",
                                         "$Account.exists()"),
    exists066      ($Account, Bool, noargs, 66, false, "exists()",
                                         "$Account.exists()"),
    count067       ($Account, Int, noargs, 67, false, "count()",
                                         "$Account.count().eq(5)"),
    id068          ($Account, $Account, noargs, 68, false, "id()",
                                         "$Account.id().exists()"),
    select069      ($Account, $Account, hasArgs, Bool, Account, 69, false, "select(",
                                         "$Account.select(t->t.toTable().exists()).exists()"),
    accNum070      ($Account, $Int, noargs, 70, true, "accNum()",
                                         "$Account.accNum().exists()"),
    av071          ($Account, $Vector, noargs, 71, true, "av()",
                                         "$Account.av().exists()"),

    //  Tuple Vector
    Vector072      (Void, Vector, variable, 72, false, "Vector",
                                         "$Vector.select(t->t.toTable().exists()).exists()"),
    toTable073     (Vector, $Vector, noargs, 73, false, "toTable()",
                                         "$Vector.select(t->t.toTable().exists()).exists()"),
    id074          (Vector, Vector, noargs, 74, false, "id()",
                                         "$Vector.select(t->t.id().toTable().exists()).exists()"),
    vfield075      (Vector, Str, field, 75, true, "vfield",
                                         "$Vector.select(t->t.vfield.equals(\"d\")).exists()"),
    vb076          (Vector, $Bank, noargs, 76, true, "vb()",
                                         "$Vector.select(t->t.vb().exists()).exists()"),
    av_077         (Vector, $Account, noargs, 77, true, "av_()",
                                         "$Vector.select(t->t.av_().exists()).exists()"),
    link2b078      (Vector, Bank, noargs, 78, false, "link2b()",
                                         "$Vector.select(t->t.link2b().toTable().exists()).exists()"),
    toBank079      (Vector, Bank, noargs, 79, false, "toBank()",
                                         "$Vector.select(t->t.toBank().toTable().exists()).exists()"),

    //  Table $Vector
    $Vector080     (Void, $Vector, variable, 80, false, "$Vector",
                                         "$Vector.exists()"),
    exists081      ($Vector, Bool, noargs, 81, false, "exists()",
                                         "$Vector.exists()"),
    count082       ($Vector, Int, noargs, 82, false, "count()",
                                         "$Vector.count().eq(5)"),
    id083          ($Vector, $Vector, noargs, 83, false, "id()",
                                         "$Vector.id().exists()"),
    select084      ($Vector, $Vector, hasArgs, Bool, Vector, 84, false, "select(",
                                         "$Vector.select(t->t.toTable().exists()).exists()"),
    vfield085      ($Vector, $Str, noargs, 85, true, "vfield()",
                                         "$Vector.vfield().exists()"),
    vb086          ($Vector, $Bank, noargs, 86, true, "vb()",
                                         "$Vector.vb().exists()"),
    av_087         ($Vector, $Account, noargs, 87, true, "av_()",
                                         "$Vector.av_().exists()"),

    //  Tuple Var
    Var000         (Void, Void, variable, 88, false, "Var",
                                         "ERROR! unknown operation -- shouldn't occur"),

    //  Tuple VarArr
    VarArr000      (Void, Void, variable, 89, false, "VarArr",
                                         "ERROR! unknown operation -- shouldn't occur"),

    //  Tuple End
    End000         (Void, Void, variable, 90, false, "End",
                                         "ERROR! unknown operation -- shouldn't occur"),
    ;

    static final int len = 3;  // # of digits after true name
  
    private final String nom;
    private final Domain input;
    private final Domain output;
    private final ArrowType arrowtype;
    private final Domain argType;
    private final boolean processArgs;
    private final Domain varType;
    private final ParseArrowType parsearrowType;
    private final int ln;
    private final boolean user;
    private String functor;    // can be updated
    private final String parsetest;

    private Arrow(Domain in, Domain out, ArrowType t, Domain argtype, Domain vartype, int ln, boolean usr, String funct, String parseTest) {
        String n = name();
        nom = n.substring(0, n.length() - len) + t.attach();
        input = in;
        output = out;
        arrowtype = t;
        argType = argtype; //Void;
        processArgs = (t == ArrowType.hasArgs);
        varType = vartype; //ull;
        switch(n) {
            case "Var000":
            case "Int000": 
            case "Str000":  parsearrowType = ParseArrowType.token; break;
            case "VarArr000": parsearrowType = ParseArrowType.tokenArrow; break;
            case "End000" : parsearrowType = ParseArrowType.end; break;
            default:
                parsearrowType = ParseArrowType.Default;
        }
        this.ln = ln;
        user = usr;
        functor = funct;
        parsetest = parseTest;
    }
    
    private Arrow(Domain in, Domain out, ArrowType t, int ln, boolean usr, String funct, String parseTest) {
        this(in,out,t,Domain.Void,null,ln,usr,funct,parseTest);
    }
    
    private Arrow(Domain in, Domain out, ArrowType t, Domain argtype,int ln,boolean usr, String funct, String parseTest) {
        this(in,out,t,argtype,null,ln,usr,funct,parseTest);
    }

    // getters
    @Override
    public String nom() {
        return nom;
    }

    @Override
    public int nomLen() {
        return nom.length();
    }

    @Override
    public Domain in() {
        return input;
    }

    @Override
    public Domain out() {
        return output;
    }

    @Override
    public ArrowType arrowType() {
        return arrowtype;
    }

    @Override
    public Domain argType() {
        return argType;
    }

    @Override
    public Domain varType() {
        return varType;
    }
    
    @Override
    public boolean processArgs() { 
        return processArgs;
    }

    @Override
    public int ln() { 
        return ln;
    }

    @Override
    public boolean user() { 
        return user;
    }

    @Override
    public String functor() { 
        return functor;
    }

    @Override
    public String getClassName() {
        String csvName = Thread.currentThread().getStackTrace()[1].getClassName() + ".csv";
        return csvName;
    }

    @Override
    public void print() {
        System.out.format("%s(%s %s)-> %s\n",name(), input, output, functor);
    }

    @Override
    public Arrow[] ArrowValues() {
        return Arrow.values();
    }

    @Override
    public ParseArrowType getParseArrowType() {
        return parsearrowType;
    }

    @Override
    public String getParseTest() {
        return parsetest;
    }

    public void set(String functorValue) {
        functor = functorValue;
    }

   public static IArrow find(String name, Domain in, Domain out) {
      for (Arrow a : Arrow.values()) {
         if (a.input == in && a.output == out) {
            String n = a.name();
            n = n.substring(0, n.length() - len);
            if (n.equals(name)) {
               return a;
            }
         }
      }
      throw new RuntimeException(String.format("Cannot find Arrow (%s,%s,%s)",name,in.toString(),out.toString()));
   }

   static { CEXCommon.BottomUpdate.update();
            CEXCommon.TopUpdate.dualUpdate();
   }

}
