package AFunctor;

import CD.AssocSchema;
import CD.CDConform;
import CD.CDSpec;
import M4.CDSpec2Png;
import PrologDB.Column;
import PrologDB.ErrorReport;
import PrologDB.Table;
import PrologDB.TableSchema;
import PrologDB.Tuple;
import XformTools.CDSpec2Apl;
import java.io.File;

public class First {

    static String cdspecEnd = ".cdspec.pl";
    static String difcsvEnd = ".diff.csv";
    static String extcsvEnd = ".extra.csv";
    static String updcsvEnd = ".update.csv";
    static String topApl = "top.apl.pl";
    static String bottomApl = "bottom.apl.pl";
    static String pngEnd = ".class.png";

    static void marquee() {
        System.err.println();
        System.err.format("Usage: %s <dirName>\n", First.class.getName());
        System.err.format("where <dirName>/<dirName>%s and\n", cdspecEnd);
        System.err.format("      <dirName>/<dirName>%s  exist\n", difcsvEnd);
        System.exit(1);
    }

    static int postit;

    private static void postInit() {
        postit = 1;
    }

    private static void post(String msg) {
        System.out.format("%d) %s\n", postit++, msg);
    }

    static Table extra, update;
    
    public static void main(String... args) throws Exception {
        postInit();
        
        // Step 1: standard marquee processing
        if (args.length != 1) {
            marquee();
        }

        // Step 2: read in cdspec and diff files
        post("reading cdspec.pl and .diff.csv files");
        String appName = args[0] + "/" + args[0];
        CDSpec cdspec = CDSpec.read(appName + cdspecEnd);
        PrologDB.DB cddb = CDSpec.toCD(cdspec);
        CDConform.check(cddb);
        Table diffcsv = Table.read(appName + difcsvEnd);
        consistencyDiff(cdspec, diffcsv);

        // Step 2: convert CDSpec2Apl
        post("converting cdspec to Top.apl.pl and Bottom.apl.pl");
        CDSpec2Apl.main(appName + cdspecEnd, appName + difcsvEnd);

        // Step 3: project CDSpecs
        post("projecting cdspec to Top.cdspec.pl and Bottom.cdspec.pl");
        CDSpec top = cdspec.project("T", diffcsv);
        CDSpec bottom = cdspec.project("B", diffcsv);
        top.print("Top.cdspec.pl");
        bottom.print("Bottom.cdspec.pl");

        // Step 4: create png files
        File pngOn = new File("PngOn");
        if (pngOn.exists()) {
            post("creating " + args[0] + ".png, top.png, and bottom.png");
            CDSpec2Png.main(appName + cdspecEnd, args[0] + pngEnd);
            CDSpec2Png.main("Top.cdspec.pl", "Top" + pngEnd);
            CDSpec2Png.main("Bottom.cdspec.pl", "Bottom" + pngEnd);
        } else {
            post("skipping creation of .png files");
        }

        // Step 5: create empty .update.csv and .extra.csv tables
        TableSchema extraTS = new TableSchema("extra").add("TorBorTB","fName","input","output","arrowtype","Notes");
        TableSchema updateTS = new TableSchema("update").add("TorB","fname","input","output","set","Notes");
        extra = new Table(extraTS);
        update = new Table(updateTS);
        
        // Step 6: create initial .update.csv and .extra.csv tables
        post("creating initial .update.csv and .extra.csv tables");
        makeTables(top,bottom);
        extra.write(args[0]+extcsvEnd);
        update.write(args[0]+updcsvEnd);
    }
    
    static void makeTables(CDSpec top, CDSpec bottom) {
        addTableLines("T",top,bottom);
        addAssocLines("T",top,bottom);
        addTableLines("B",bottom,top);
        addAssocLines("B",bottom,top);
    }

    
    private static void addTableLines(String TorB,CDSpec a, CDSpec b) {
        for (TableSchema ts : a.getTableSchemas()) { 
            String tabName = "$"+ts.getName();
            // add a row to the update.csv for each table in a that is not in b
            TableSchema bts = b.getTableSchema(ts.getName());
            if (bts==null) {
                update.add(TorB,tabName,"Void",tabName,"xxx","");
                // add a row in update.csv for each field of this table
                for (Column c : ts.getColumns()) {
                    String cname = c.getName();
                    String kname = c.getKind();
                    String Kname = kname.substring(0,1).toUpperCase()+kname.substring(1);
                    update.add(TorB,cname,tabName.substring(1),Kname,"xxx","");
                    update.add(TorB,cname,tabName,"$"+Kname,"xxx","");
                }
            }  else {
                // add a row in update.csv for each field in the shared table of a that is not in b
                for (Column c : ts.getColumns()) {
                    String cname = c.getName();
                    String kname = c.getKind();
                    String Kname = kname.substring(0,1).toUpperCase()+kname.substring(1);
                    if (bts.getColumn(cname)==null) {
                        update.add(TorB,cname,tabName.substring(1),Kname,"xxx","");
                        update.add(TorB,cname,tabName,"$"+Kname,"xxx","");
                    }
                }
            } 
        }
        
    }
    
    private static void addAssocLines(String TorB, CDSpec a, CDSpec b) {
        for (AssocSchema as : a.getAssocSchemas()) {
            String asName = as.getName();
            AssocSchema bas = b.getAssocSchema(asName);
            if (bas==null) {
                String leftName = as.getClass1();
                String rightName = as.getClass2();
                update.add(TorB,asName,rightName, "$"+leftName,"xxx","");
                update.add(TorB,asName,"$"+rightName, "$"+leftName,"xxx","");
                if (as.getMin1()==1 && as.getMax1()==1) {
                    update.add(TorB, "link"+asName,rightName,leftName,"xxx","");
                    extra.add(TorB,"link"+asName,rightName,leftName,"noargs","" );
                }
                update.add(TorB,asName+"_",leftName, "$"+rightName,"xxx","");
                update.add(TorB,asName+"_","$"+leftName, "$"+rightName,"xxx","");
                if (as.getMin2()==1 && as.getMax2()==1) {
                    update.add(TorB, "link"+asName+"_",leftName,rightName,"xxx","");
                    extra.add(TorB,"link"+asName+"_",leftName,rightName,"noargs","" );
                }
            }
        }
    }
    
    
    static void consistencyDiff(CDSpec cdspec, Table diffcsv) {
        ErrorReport er = new ErrorReport();
        for (Tuple t : diffcsv.tuples()) {
            String acname = t.get("name");
            switch (t.get("CorA")) {
                case "C":
                    TableSchema ts = cdspec.getTableSchema(acname);
                    if (ts == null) {
                        er.add("reference to Class (%s) in .diff.csv does not exist in .cspec.pl file",acname);
                    }
                    String field = t.get("field");
                    if (field.equals("")) break;
                    Column c = ts.getColumn(field);
                    if (c == null) {
                        er.add("reference to Class.Field (%s.%s) in .diff.csv does not exist in .cspec.pl file",acname,field);
                    }
                    
                    break;
                case "A":
                    if (cdspec.getAssocSchema(acname) == null) {
                        er.add("reference to Association (%s) in .diff.csv does not exist in .cspec.pl file",acname);
                    }
                    break;
                default:
                    er.add("unknown CorA value = '" + t.get("CorA") + "' in .diff.csv file in .cspec.pl file");
            }
        }
        er.printEH();
    }
    
    static void consistencyUpd(CDSpec cdspec, Table ucsv) {
        ErrorReport er = new ErrorReport();
        String u = "in .update.csv file";
        for (Tuple t: ucsv.tuples()) {
            String TorB = t.get("TorB");
            if (!(TorB.equals("T") || TorB.equals("B"))) {
                er.add("%s tuple %s, TorB column is not 'T' or 'B'",u,t.toString());
                continue;
            }
            consistencyTable(cdspec,t.get("input"),t,er,u);
            consistencyTable(cdspec,t.get("output"),t,er,u);
        }
        er.printEH();
    }
    
    static void consistencyTable(CDSpec cdspec, String tname, Tuple t, ErrorReport er, String u) {
        if (tname.equals("Void"))
            return;
        if (tname.startsWith("$"))
            tname = tname.substring(1);
        if (tname.equals("Str") || tname.equals("Int") || tname.equals("Bool") || tname.equals("Pair")) 
            return;
        if (cdspec.getTableSchema(tname)==null) {
                er.add("%s tuple %s: non-existent Tuple or Table %s type referenced",u,t.toString(),tname);
        }
    }
    
    static void consistencyExtra(CDSpec cdspec, Table xcsv) {
        ErrorReport er = new ErrorReport();
        String u = "in .extra.csv file";
        for (Tuple t: xcsv.tuples()) {
            String TorB = t.get("TorBorTB");
            if (!(TorB.equals("T") || TorB.equals("B") || TorB.equals("TB"))) {
                String tpl = t.toString();
                er.add("%s tuple %s: TorBorTB column is not 'T' or 'B' or 'TB'",u,tpl);
                continue;
            }
            consistencyTable(cdspec,t.get("input"),t,er,u);
            consistencyTable(cdspec,t.get("output"),t,er,u);
            
            String arrowtype = t.get("arrowtype");
            if (!(arrowtype.equals("noargs") || arrowtype.equals("field"))) {
                er.add("%s tuple %s:, arrowtype column is not 'noargs' or 'field'",u,t.toString());
            }
            
            er.printEH();
        }
    }
}
