package AFunctor;

import CD.CDSpec;
import M4.CDSpec2Png;
import PrologDB.Column;
import PrologDB.ErrorReport;
import PrologDB.Table;
import PrologDB.TableSchema;
import PrologDB.Tuple;
import XformTools.Apl2Fpl;
import XformTools.CDSpec2Apl;
import java.io.File;

public class MainOld {

    static String cdspecEnd = ".cdspec.pl";
    static String difcsvEnd = ".diff.csv";
    static String extcsvEnd = ".extra.csv";
    static String updcsvEnd = ".update.csv";
    static String topApl = "top.apl.pl";
    static String bottomApl = "bottom.apl.pl";
    static String pngEnd = ".class.png";

    static void marquee() {
        System.err.println();
        System.err.format("Usage: %s <applicationName>\n", MainOld.class.getName());
        System.err.format("where <applicationName>%s and\n", cdspecEnd);
        System.err.format("      <applicationName>%s  and\n", difcsvEnd);
        System.err.format("      <applicationName>%s  exist\n", extcsvEnd);
        System.exit(1);
    }

    static int postit;

    private static void postInit() {
        postit = 1;
    }

    private static void post(String msg) {
        System.out.format("%d) %s\n", postit++, msg);
    }
    public static int warningCount = 0;   // referenced only in JArrow

    public static void main(String... args) throws Exception {
        postInit();
        // Step 1: standard marquee processing
        if (args.length != 1) {
            marquee();
        }
        String appName = args[0] + "/" + args[0];

        post("reading cdspec.pl, .diff.csv, .extra.csv, and .update.csv files");
        CDSpec cdspec = CDSpec.read(appName + cdspecEnd);
        Table diffcsv = Table.read(appName + difcsvEnd);
        String XcsvName = appName + extcsvEnd;
        Table xcsv = Table.read(XcsvName);  // make sure no errors in read
        String UcsvName = appName + updcsvEnd;
        Table ucsv = Table.read(UcsvName);// make sure no errors in read

        consistencyDiff(cdspec, diffcsv);
        consistencyExtra(cdspec,xcsv);
        consistencyUpd(cdspec, ucsv);

        // Step 2: convert CDSpec2Apl
        post("converting cdspec to Top.apl.pl and Bottom.apl.pl");
        CDSpec2Apl.main(appName + cdspecEnd, appName + difcsvEnd);

        // Step 3: project CDSpecs
        post("projecting cdspec to Top.cdspec.pl and Bottom.cdspec.pl");
        CDSpec top = cdspec.project("T", diffcsv);
        CDSpec bottom = cdspec.project("B", diffcsv);
        top.print("Top.cdspec.pl");
        bottom.print("Bottom.cdspec.pl");

        // Step 4: create png files
        File pngOn = new File("PngOn");
        if (pngOn.exists()) {
            post("creating " + args[0] + ".png, top.png, and bottom.png");
            CDSpec2Png.main(appName + cdspecEnd, args[0] + pngEnd);
            CDSpec2Png.main("Top.cdspec.pl", "Top" + pngEnd);
            CDSpec2Png.main("Bottom.cdspec.pl", "Bottom" + pngEnd);
        } else {
            post("skipping creation of .png files");
        }

        // Step 5: convert apl files to fpl files
        post("transforming Top.apl.pl -> Top.fpl.pl and Bottom.apl.pl -> Bottom.apl.pl");
        Apl2Fpl.top("Top.apl.pl", XcsvName, "Top.fpl.pl");
        Apl2Fpl.bot("Bottom.apl.pl", XcsvName, "Bottom.fpl.pl");

        // Step 6: now run JRefact (or Apl2Java)
        post("converting apl files to Java");
        warningCount = 0;
        jrefact.JRefact.main("Top.fpl.pl","Bottom.fpl.pl", xcsv, ucsv, diffcsv);
        System.out.println();

        post("look in ABET4/AAOutput");
    }

    static void consistencyDiff(CDSpec cdspec, Table diffcsv) {
        ErrorReport er = new ErrorReport();
        for (Tuple t : diffcsv.tuples()) {
            String acname = t.get("name");
            switch (t.get("CorA")) {
                case "C":
                    TableSchema ts = cdspec.getTableSchema(acname);
                    if (ts == null) {
                        er.add("reference to Class (%s) in .diff.csv does not exist in .cspec.pl file",acname);
                    }
                    String field = t.get("field");
                    if (field.equals("")) break;
                    Column c = ts.getColumn(field);
                    if (c == null) {
                        er.add("reference to Class.Field (%s.%s) in .diff.csv does not exist in .cspec.pl file",acname,field);
                    }
                    
                    break;
                case "A":
                    if (cdspec.getAssocSchema(acname) == null) {
                        er.add("reference to Association (%s) in .diff.csv does not exist in .cspec.pl file",acname);
                    }
                    break;
                default:
                    er.add("unknown CorA value = '" + t.get("CorA") + "' in .diff.csv file in .cspec.pl file");
            }
        }
        er.printEH();
    }
    
    static void consistencyUpd(CDSpec cdspec, Table ucsv) {
        ErrorReport er = new ErrorReport();
        String u = "in .update.csv file";
        for (Tuple t: ucsv.tuples()) {
            String TorB = t.get("TorB");
            if (!(TorB.equals("T") || TorB.equals("B"))) {
                er.add("%s tuple %s, TorB column is not 'T' or 'B'",u,t.toString());
                continue;
            }
            consistencyTable(cdspec,t.get("input"),t,er,u);
            consistencyTable(cdspec,t.get("output"),t,er,u);
        }
        er.printEH();
    }
    
    static void consistencyTable(CDSpec cdspec, String tname, Tuple t, ErrorReport er, String u) {
        if (tname.equals("Void"))
            return;
        if (tname.startsWith("$"))
            tname = tname.substring(1);
        if (tname.equals("Str") || tname.equals("Int") || tname.equals("Bool") || tname.equals("Pair")) 
            return;
        if (cdspec.getTableSchema(tname)==null) {
                er.add("%s tuple %s: non-existent Tuple or Table %s type referenced",u,t.toString(),tname);
        }
    }
    
    static void consistencyExtra(CDSpec cdspec, Table xcsv) {
        ErrorReport er = new ErrorReport();
        String u = "in .extra.csv file";
        for (Tuple t: xcsv.tuples()) {
            String TorB = t.get("TorBorTB");
            if (!(TorB.equals("T") || TorB.equals("B") || TorB.equals("TB"))) {
                String tpl = t.toString();
                er.add("%s tuple %s: TorBorTB column is not 'T' or 'B' or 'TB'",u,tpl);
                continue;
            }
            consistencyTable(cdspec,t.get("input"),t,er,u);
            consistencyTable(cdspec,t.get("output"),t,er,u);
            
            String arrowtype = t.get("arrowtype");
            if (!(arrowtype.equals("noargs") || arrowtype.equals("field"))) {
                er.add("%s tuple %s:, arrowtype column is not 'noargs' or 'field'",u,t.toString());
            }
            
            er.printEH();
        }
    }
}
