package AFunctor;

import CD.CDSpec;
import PrologDB.ErrorReport;
import PrologDB.Table;
import PrologDB.Tuple;
import XformTools.Apl2Fpl;

public class Second {

    static String cdspecEnd = ".cdspec.pl";
    static String difcsvEnd = ".diff.csv";
    static String extcsvEnd = ".extra.csv";
    static String updcsvEnd = ".update.csv";
    static String topApl = "top.apl.pl";
    static String bottomApl = "bottom.apl.pl";
    static String pngEnd = ".class.png";

    static void marquee() {
        System.err.println();
        System.err.format("Usage: %s <applicationName>\n", Second.class.getName());
        System.err.format("where <dirName>/<dirName>%s and\n", cdspecEnd);
        System.err.format("      <dirName>/<dirName>%s  exist\n", difcsvEnd);
        System.exit(1);
    }

    static int postit;

    private static void postInit() {
        postit = 6;
    }

    private static void post(String msg) {
        System.out.format("%d) %s\n", postit++, msg);
    }
    public static int warningCount = 0;   // referenced only in JArrow

    public static void main(String... args) throws Exception {
        postInit();
        // Step 1: standard marquee processing
        if (args.length != 1) {
            marquee();
        }
        String appName = args[0] + "/" + args[0];

        post("reading cdspec.pl, .diff.csv, .extra.csv and .update.csv files");
        CDSpec cdspec = CDSpec.read(appName + cdspecEnd);
        Table diffcsv = Table.read(appName + difcsvEnd);
        String XcsvName = appName + extcsvEnd;
        Table xcsv = Table.read(XcsvName);  // make sure no errors in read
        String UcsvName = appName + updcsvEnd;
        Table ucsv = Table.read(UcsvName);// make sure no errors in read
        consistencyExtra(cdspec, xcsv);


        // Step 2: convert apl files to fpl files
        post("transforming Top.apl.pl -> Top.fpl.pl and Bottom.apl.pl -> Bottom.apl.pl");
        Apl2Fpl.top("Top.apl.pl", XcsvName, "Top.fpl.pl");
        Apl2Fpl.bot("Bottom.apl.pl", XcsvName, "Bottom.fpl.pl");
        consistencyUpd(cdspec, ucsv);

        // Step 3: now run JRefact (or Apl2Java)
        post("converting apl files to Java");
        warningCount = 0;
        jrefact.JRefact.main("Top.fpl.pl", "Bottom.fpl.pl", xcsv, ucsv, diffcsv);
        System.out.println();

        post("look in AAOutput");
    }

    static void consistencyUpd(CDSpec cdspec, Table ucsv) {
        ErrorReport er = new ErrorReport();
        String u = "in .update.csv file";
        for (Tuple t : ucsv.tuples()) {
            String TorB = t.get("TorB");
            if (!(TorB.equals("T") || TorB.equals("B"))) {
                er.add("%s tuple %s, TorB column is not 'T' or 'B'", u, t.toString());
                continue;
            }
            consistencyTable(cdspec, t.get("input"), t, er, u);
            consistencyTable(cdspec, t.get("output"), t, er, u);
        }
        consistencySet(ucsv, er);
        er.printEH();
    }

    static void consistencySet(Table ucsv, ErrorReport er) {
        // Step 1: examine every set value and see if it is defined
        boolean xxxSeen = false, firstAdd = true, bigjump = false;
        
        for (Tuple t : ucsv.tuples()) {
            String set = t.get("set").replace("()", "");
            if (set.equals("xxx")) {
                xxxSeen = true;
                continue;
            }
            String[] sets = set.split("\\.");
            String BorT = t.get("TorB").equals("T") ? "B" : "T";
            Tuple tup = t;
            String input = t.get("input");
            for (int i = 0; i < sets.length; i++) {
                String fname = sets[i];
                if (input.equals("Void")) {
                    System.out.format("warning: manually verify there is a table domain %s\n", fname);
                    input = fname;
                    continue;
                }
                if (input.equals("Pair") || input.equals("$Pair")) {
                    bigjump = true;
                    break;
                }
                tup = find(ucsv, BorT, fname, input);
                if (tup == null) {
                    if (firstAdd) {
                        System.out.println("\nadd and complete the following tuple(s) to .update.csv");
                        firstAdd = false;
                    }
                    er.add("%s,%s,%s,%s,%s,%s\n", BorT, fname, input, "???", "xxx"," (*could not verify existence*)");
                    bigjump = true;
                    break;
                }
                input = tup.get("output");
            }
            if (!bigjump && !(t.get("output").equals(tup.get("output")))) {
                er.add("suspected set type mismatch in tuple : %s", t.toString());
            }
            bigjump = false;
        }
        if (xxxSeen) {
            er.add("remove all set('xxx') in .update.csv");
        }
        System.out.println();
    }

    static Tuple find(Table ucsv, String TorB, String fname, String input) {
        Tuple t = ucsv.getFirst(u -> u.is("TorB", TorB) && u.is("fname", fname) && u.is("input", input));
        return t;
    }

    static void consistencyTable(CDSpec cdspec, String tname, Tuple t, ErrorReport er, String u) {
        if (tname.equals("Void")) {
            return;
        }
        if (tname.startsWith("$")) {
            tname = tname.substring(1);
        }
        if (tname.equals("Str") || tname.equals("Int") || tname.equals("Bool") || tname.equals("Pair")) {
            return;
        }
        if (cdspec.getTableSchema(tname) == null) {
            er.add("%s tuple %s: non-existent Tuple or Table %s type referenced", u, t.toString(), tname);
        }
    }

    static void consistencyExtra(CDSpec cdspec, Table xcsv) {
        ErrorReport er = new ErrorReport();
        String u = "in .extra.csv file";
        for (Tuple t : xcsv.tuples()) {
            String TorB = t.get("TorBorTB");
            if (!(TorB.equals("T") || TorB.equals("B") || TorB.equals("TB"))) {
                String tpl = t.toString();
                er.add("%s tuple %s: TorBorTB column is not 'T' or 'B' or 'TB'", u, tpl);
                continue;
            }
            consistencyTable(cdspec, t.get("input"), t, er, u);
            consistencyTable(cdspec, t.get("output"), t, er, u);

            String arrowtype = t.get("arrowtype");
            if (!(arrowtype.equals("noargs") || arrowtype.equals("field"))) {
                er.add("%s tuple %s:, arrowtype column is not 'noargs' or 'field'", u, t.toString());
            }

            er.printEH();
        }
    }
}
