package Common;

import Parsing.LineParsers.PrimitiveLookAhead;
//import PrologDB.Table;
//import PrologDB.Tuple;
//import java.io.File;
//import java.io.PrintStream;
//import java.util.HashMap;
import java.util.LinkedList;
import java.util.Stack;

public abstract class GSentence<
        Domain extends IDomain, Arrow extends IArrow<Arrow, Domain>>
        implements ISentence {

    static protected PrimitiveLookAhead pla = new PrimitiveLookAhead();

    abstract public Domain getVoid();

    abstract public Domain getBool();

    abstract public Domain getInt();

    abstract public Domain getStr();

    abstract public Domain[] values();

    abstract public GSentence New(String s);

    abstract public Arrow getEnd000();

    abstract public Arrow getInt000();

    abstract public Arrow getStr000();

    abstract public Arrow getVarArr000();

    abstract public Arrow getVar000();

    abstract public Arrow[] arrowValues();
    
    abstract public String getClassName();

    protected String toParseOriginal;     // original input string
    protected String toParse;             // what remains to be parsed
    private final LinkedList<Arrow> list;       // this is the parsed output
    protected Domain domc;                // current domain (type) at toParse
    protected Domain domr;                // required domain(type) at end of parse
    protected Stack<Domain2> stack;       // used for pushing contexts
    protected LinkedList<String> tokens;  // integer & string tokens that were parsed
    public Stack<Vars> vars;           // stack of single-letter variables from select statements
//    public HashMap<String, String> functor;

    // getters
    public String toParse() {
        return toParse;
    }

    public char firstChar() {
        return toParse.charAt(0);
    }

    public String origSen() {
        return toParseOriginal;
    }

    public Domain rdom() {
        return domr;
    }

    public Domain cdom() {
        return domc;
    }

    public void setCdom(Domain d) {
        domc = d;
    }

    public LinkedList<Arrow> parsedList() {
        return list;
    }

//    private HashMap<String, String> getFunctor() {
//        if (functor == null) {
//            functor = ReadFunctorFile();
//        }
//        return functor;
//    }

    // this a kind of reset; reset sentence to process toParse
    // to Domain.void and save the current parsed output.
    public void push(Domain d) {
        Domain2 d2 = new Domain2(domc, domr);
        stack.push(d2);
        domc = getVoid();
        domr = d;
    }

    public GSentence(String s) {
        toParseOriginal = s;
        toParse = removeBlanks(s);
        list = new LinkedList<>();
        domr = getBool();
        domc = getVoid();
        stack = new Stack<>();
        tokens = new LinkedList<>();
        vars = new Stack<>();
////        functor = null;
    }

    // ifparse -- consume pairs
    ///////////////////////// next char of toParse is )?
    public boolean endExpr() {
        return toParse.startsWith(")");
    }

    public void parseExpr() {
        // we have come to the end of an expression.
        // fix toParse to continue and make sure that the current domain(type)
        // matches what is expected, otherwise an argument was provided
        // that is not the right type.
        String tp = toParse.substring(1);
        toParse = tp;
        if (domc != domr) {
            throw new RuntimeException("Argument type mismatch before " + toParse);
        }

        // Now continue with the conditions from where we left off
        Domain2<Domain> d2 = stack.pop();
        domc = d2.current();
        domr = d2.required();
        list.add(getEnd000());
    }

    /////////////////////////  next chars of toParse is an arrow name?
    public boolean startsWith(Arrow a) {
        return toParse.startsWith(a.nom());
    }

    public void parse(Arrow a) {
        String tp = toParse.substring(a.nomLen());
        toParse = tp;
        domc = a.out();
        list.add(a);
    }

    ///////////////////// have we consumed the last character?
    public boolean atEnd() {
        boolean result = (toParse.length() == 0);
        return result;
    }

    //////////////////////// is next character .?
    public boolean isDot() {
        return toParse.startsWith(".");
    }

    public void parseDot() {
        toParse = toParse.substring(1);
    }

    //////////////////////////// is next char ;?
    public boolean isSemiColon() {
        return toParse.startsWith(";");
    }

    public void parseSemiColon() {
        if (domc != domr) {
            throw new RuntimeException("statement does not type check to Bool");
        }
        int ss = stack.size();
        if (ss != 0) {
            throw new RuntimeException("incomplete expression as end of line reached");
        }
        toParse = "";
    }

    /////////////////////////// is next token an (signed)integer?
    public boolean isInt() {
        return pla.isNumber(toParse);
    }

    public void parseInt() {
        char c = toParse.charAt(0);
        int i = (c == '-') ? 1 : 0;
        while (true) {
            i++;
            char k = toParse.charAt(i);
            if (pla.isDigit(k)) {
                continue;
            }
            break;
        }
        String parsed = toParse.substring(0, i);
        toParse = toParse.substring(i);
        domc = getInt();
        tokens.add(parsed);
        list.add(getInt000());
    }

    /////////////////////////// is next token an (signed)integer?
    public boolean isQuotedString() {
        char c = toParse.charAt(0);
        return pla.isQuote(c);
    }

    public void parseQuotedString() {
        char c = toParse.charAt(0);
        int i = 0;
        while (true) {
            i++;
            char k = toParse.charAt(i);
            if (k == c) {
                break;
            }
        }
        String parsed = toParse.substring(0, i + 1);
        toParse = toParse.substring(i + 1);
        domc = getStr();
        tokens.add(parsed);
        list.add(getStr000());
    }

    // check if 'x->' start of a select statement
    public boolean isVarArrow() {
        char c = toParse.charAt(0);
        if (!pla.isLetter(c)) {
            return false;
        }
        String ca = c + "->";
        return toParse.startsWith(ca);
    }

    public void parseVarArrow() {
        char c = toParse.charAt(0);
        toParse = toParse.substring(3);
        domc = getVoid();

        Arrow a = list.getLast();
        if (!a.nom().equals("select(")) {
            throw new RuntimeException("Expected select operation prior to 't->' parse");
        }
        Domain vartype = a.varType();
        Vars v = new Vars(c, vartype);
        vars.push(v);
        list.add(getVarArr000());
        tokens.add(String.valueOf(c));
    }

    //////////////////////////////////
    public boolean isVar() {
        char c = toParse.charAt(0);
        char p = toParse.charAt(1);
        if (p != '.') // must be something following a variable
        {
            return false;
        }
        for (Vars v : vars) {
            if (v.name() == c) {
                handoff = v;
                return true;
            }
        }
        return false;
    }

    private Vars<Domain> handoff = null;

    public void parseVar() {
        String tp = toParse.substring(1); //keep the dot!!
        toParse = tp;
        domc = handoff.dom();
        tokens.add(String.valueOf(handoff.name()));
        list.add(getVar000());
    }

    /////////////////////////////// printing utilities
    public String toString() {
        int i = 0;
        String result = "";
        for (Arrow a : list) {
            ArrowType at = a.arrowType();
            String add = (at == ArrowType.variable) ? "" : ".";
            ParseArrowType pat = a.getParseArrowType();
            switch (pat) {
                case token:
                    result = result + tokens.get(i++);
                    break;
                case tokenArrow:
                    result = result + tokens.get(i++) + "->";
                    break;
                case end:
                    result = result + ")";
                    break;
                default:
                    result = result + add + a.nom();
                    break;
            }
        }
        result = result + ";";
        return result;
    }

    public String Translate() {
//        HashMap<String, String> functorHM = getFunctor(); /// ReadFunctorFile();
        int i = 0;
        int j;
        String result = "", repl;
        for (Arrow a : list) {
            String add = (a.arrowType() == ArrowType.variable) ? "" : ".";
            switch (a.getParseArrowType()) {
                case token:
                    result = result + tokens.get(i++);
                    break;
                case tokenArrow:
                    result = result + tokens.get(i++) + "->";
                    break;
                case end:
                    result = result + ")";
                    break;
                default:
//                    String key = a.name();
//                    repl = functorHM.get(key);
                    repl = a.functor();
                    if (repl == null) {
                        throw new RuntimeException("null translation for " + a.name());
                    }
                    result = result + add + repl;
                    break;
            }
        }
        result = result + ";";
        return result;
    }

    public void printParse() {
        System.out.format("input: %s\n", toParseOriginal);
        System.out.println(this.toString());
    }

    ///// remove blanks
    public static String removeBlanks(String s) {
        char[] c = s.toCharArray();
        String n = "";
        boolean insideString = false;
        for (int i = 0; i < c.length; i++) {
            char k = c[i];
            if (k != ' ' || insideString) {
                n += k;
            }
            if (c[i] == '\"') {
                insideString = !insideString;
            }
        }
        return n;
    }

    public void printList() {
        System.out.println("Derivation:");
        for (Arrow a : list) {
            a.print();
        }
    }

    /// experimental
    public LinkedList<Arrow> getArrows(Domain in) {
        LinkedList<Arrow> list = new LinkedList<>();
        for (Arrow a : arrowValues()) {
            if (a.in() == in) {
                list.add(a);
            }
        }
        return list;
    }

    public LinkedList<Arrow> getArrowsThatStart() {
        LinkedList<Arrow> list = new LinkedList<>();
        for (Arrow a : arrowValues()) {
            if (a.in() == getVoid()) {
                if (a.out().start()) {
                    list.add(a);
                }
            }
        }
        return list;
    }

    //DON: create Interface for Domain (to access start)
    // parsing routine.  Sentence s is input.
    public void Parse() {
        // Step 1: can't be at the end of a line;
        //         end of line occurs when ; is seen as the last (non-white) 
        //         character
        if (atEnd()) {
            throw new RuntimeException("unexpected end of sentence");
        }

        // see if we are at the end of a parse of an arrow with args...
        if (endExpr()) {
            parseExpr();
            return;
        }

        // Step 2: process void arrows differently
        if (cdom() == getVoid()) {
            for (Arrow a : getArrowsThatStart()) {
                if (startsWith(a)) {
                    parse(a);
                    Parse();
                    return;
                }
            }
            if (isInt()) {
                parseInt();
                Parse();
                return;
            }
            if (isQuotedString()) {
                parseQuotedString();
                Parse();
                return;
            }
            if (isVarArrow()) {
                parseVarArrow();
                Parse();
                return;
            }
            if (isVar()) {
                parseVar();
                Parse();
                return;
            }
            throw new RuntimeException("unexpected start (lead token) of sentence");
        }

        // Step 3: toParse must now begin with . or ; 
        if (isSemiColon()) {
            parseSemiColon();
            return;   // unroll stack-- done parsing sentence
        }
        if (isDot()) {
            parseDot();
            // continue with arrow parsing
        } else {
            throw new RuntimeException("Could not parse " + toParse());
        }
        for (Arrow a : getArrows(cdom())) {
            if (startsWith(a)) {
                parse(a);
                if (a.processArgs()) {
                    push(a.argType()); // resets s.dom to Void
                    Parse();// parse the argument
                    if (a.varType() != null) { // this is select-like operation
                        vars.pop().name(); // pop off stack variable that was defined xxxxxxxxx
                    }
                }
                Parse();
                return;
            }
        }

        throw new RuntimeException("Could not parse " + toParse());
    }

//    public void main000(String... args) throws Exception {
//
//        String sentenceName = getClassName();
//        String functorFile = sentenceName.replace("Sentence", "Functor") + ".csv";
//        String RoundTripFile = functorFile.replace("Functor", "RoundTrip");
//
//        cannotExist(functorFile);
//        cannotExist(RoundTripFile);
//        writeFile(functorFile, true);
//        writeFile(RoundTripFile, false);
//    }

//    private void cannotExist(String filename) {
//        File oldfile = new File(filename);
//        if (oldfile.exists()) {
//            System.err.format("You must move %s to backup!\n", filename);
//            System.exit(1);
//        }
//    }
//
//    private void writeFile(String filename, boolean isFunctorFile) throws Exception {
//        PrintStream ps = new PrintStream(filename);
//        ps.format("Translate,name,input,output,arrowtype,ln,user\n");
//        for (Arrow a : arrowValues()) {
//            String firstcol = isFunctorFile ? a.nom() : mkTrans(a);
//            ps.format("%s,%s,%s,%s,%s,%d,%b\n",
//                    firstcol, a.name(), a.in().name(), a.out().name(), a.arrowType().name(), a.ln(), a.user());
//        }
//        ps.close();
//        System.out.println(filename + " created!");
//    }
//
//    private String mkTrans(Arrow a) {
//        // skip over standard operations
//        String name = a.name();
//        String number = name.substring(name.length() - 3);
//        int n = Integer.parseInt(number);
//        if (n < 10) {
//            return "";
//        }
//
//        ////////////////////////////////////  VARIABLE //////////////////////////////
//        if (a.arrowType() == ArrowType.variable) {
//            if (a.in().name().equals("Void")) {
//                if (a.out().name().startsWith("$")) {
//                    return String.format("%s.exists()", a.out().name());
//                } else {
//                    return String.format("$%s.select(t->t.toTable().exists()).exists()", a.out().name());
//                }
//            }
//            if (a.in().name().startsWith("$")) {
//                switch (a.out().name()) {
//                    case "Str":
//                        return String.format("%s.select(t->t.%s.equals(\"d\")).exists()", a.in().name(), a.nom());
//                    case "Int":
//                        return String.format("%s.select(t->t.%s.eq(2)).exists()", a.in().name(), a.nom());
//                    case "Bool":
//                        return String.format("%s.select(t->t.%s).exists()", a.in().name(), a.nom());
//                    default:
//                        throw new RuntimeException("unknown output " + a.out().name());
//                }
//            } else {
//                switch (a.out().name()) {
//                    case "Str":
//                        return String.format("$%s.select(t->t.%s.equals(\"d\")).exists()", a.in().name(), a.nom());
//                    case "Int":
//                        return String.format("$%s.select(t->t.%s.eq(2)).exists()", a.in().name(), a.nom());
//                    case "Bool":
//                        return String.format("$%s.select(t->t.%s).exists()", a.in().name(), a.nom());
//                    default:
//                        throw new RuntimeException("unknown output " + a.out().name());
//                }
//            }
//        }
//
//        ////////////////////////////////////  NOARGS //////////////////////////////
//        if (a.arrowType() == ArrowType.noargs) {
//            if (a.in().name().startsWith("$")) {
//                switch (a.out().name()) {
//                    case "Str":
//                        return String.format("%s.%s.equals(\"d\")", a.in().name(), a.nom());
//                    case "Int":
//                        return String.format("%s.%s.eq(5)", a.in().name(), a.nom());
//                    case "Bool":
//                        return String.format("%s.%s", a.in().name(), a.nom());
//                    default:
//                        if (a.out().name().startsWith("$")) {
//                            return String.format("%s.%s.exists()", a.in().name(), a.nom());
//                        } else {
//                            return String.format("%s.select(t->t.%s.toTable().exists()).exists()", a.in().name(), a.nom());
//                        }
//                }
//            } else {
//                if (a.out().name().startsWith("$")) {
//                    return String.format("$%s.select(t->t.%s.exists()).exists()", a.in().name(), a.nom());
//                } else {
//                    return String.format("$%s.count().eq(0)", a.in().name(), a.nom());
//                }
//            }
//        }
//
//        ////////////////////////////////////  HASARGS //////////////////////////////
//        if (a.arrowType() == ArrowType.hasArgs) {
//            if (a.in().name().startsWith("$") && a.out().name().startsWith("$")) {
//                return String.format("%s.%st->t.toTable().exists()).exists()", a.in().name(), a.nom());
//            } else {
//                return String.format("$%s.select(t->t.%s.exists()).exists()", a.in().name(), a.nom());
//            }
//        }
//
//        ////////////////////////////////////  Field //////////////////////////////
//        if (a.arrowType() == ArrowType.field) {
//            switch (a.out().name()) {
//                case "Str":
//                        return String.format("$%s.select(t->t.%s.equals(\"d\")).exists()", a.in().name(), a.nom());
//                case "Int":
//                        return String.format("$%s.select(t->t.%s.eq(5)).exists()", a.in().name(), a.nom());
//                case "Bool":
//                        return String.format("$%s.select(t->t.%s).exists()", a.in().name(), a.nom());
//                default:
//                        throw  new RuntimeException("not primitive type: " +a.out().name());
//            }
//        }
//            
//        ////////////////////////////////////  End //////////////////////////////  
//        return "";
//    }
//
//    public HashMap<String, String> ReadArrowCSVFile() {
//        String cn = getClassName() + ".csv";
//        return null;
//    }
//
//    public HashMap<String, String> ReadRoundTripFile() {
//        String cn = getClassName() + ".csv";
//        cn = cn.replace("Sentence", "RoundTrip");
//        return ReadCSV(cn);
//    }
//
//    public HashMap<String, String> ReadFunctorFile() {
//        String cn = getClassName() + ".csv";
//        cn = cn.replace("Sentence", "Functor");
//        return ReadCSV(cn);
//    }
//
//    public HashMap<String, String> ReadCSV(String filename) {
//        Table csv = Table.read(filename);
//        HashMap<String, String> hashmap = new HashMap<>();
//        for (Tuple t : csv.tuples()) {
//            String key = t.get("name");
//            String value = t.get("Translate");
//
//            // next few lines are a complete hack
//            // ms .csv files have quirks... when entries have strings
//            if (value.startsWith("\"")) {
//                value = value.substring(1, value.length() - 1);
//            }
//            value = value.replace("\"\"", "\"");
//
//            if (value == null || value.equals("")) {
//                continue;
//            }
//            if (hashmap.get(key) == null) {
//                hashmap.put(key, value);
//            } else {
//                System.err.format("Duplicate entries for %s\n", key);
//            }
//        }
//
//        return hashmap;
//    }
}
