package Common;

import PrologDB.Tuple;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import static org.junit.Assert.fail;

public abstract class GenericRoundTrip<
        StartSentence extends ISentence, MidSentence extends ISentence> {

    public StartSentence startSentenceFactory;
    public MidSentence midSentenceFactory;

    public GenericRoundTrip(StartSentence startFactory, MidSentence midFactory) {
        startSentenceFactory = startFactory;
        midSentenceFactory = midFactory;
    }

    public void roundTrip(IArrow[] arr) {
        boolean succeeded = true;
        for (IArrow a : arr) {
            String e = a.getParseTest() + ";";
            if (e.startsWith("ERROR")) {
                continue;
            }
            try {
            succeeded = onetest(e,a.name()) && succeeded;
            } catch (Exception x) { 
                succeeded = false;
            }
        }
        if (!succeeded)
            fail();
    }

    
    public boolean onetest(String e, String name) {
        StartSentence s0;
        String ss;
        MidSentence s1 = null;
        String end = null;
        String start = e;
        s0 = (StartSentence) startSentenceFactory.New(start);
        try {
            s0.Parse(); // startSentenceFactory.Parse(s0);
            String e1 = s0.toString();
            boolean eeSame = (e1.equals(e));   /// for debugging
            ss = s0.Translate();
        } catch (Exception x) {
            System.err.format("to-direction failed (%s) msg = %s\n" +
                    "          toParse = %s\n\n",
                    e, x.getMessage(), s0.toParse());
            return false;
        }
        try {
            s1 = (MidSentence) midSentenceFactory.New(ss);
            s1.Parse(); //midSentenceFactory.Parse(s1);
            String sss = s1.toString();
            boolean same = (sss.equals(ss));    /// for debugging
            end = s1.Translate();
        } catch (Exception x) {
            System.err.format("on-the-way back failed: [%s]=>\n"+
                              "                        [%s]=>\n"+
                              "                        [%s]\n" +
                              "             msg = %s\n"+ 
                              "         toParse = %s\n\n", e, ss, end, x.getMessage(), s1.toParse());          
            return false;
        }

        String start1 = start; //.replace(".exists();", "");
        String end1 = end; //.replace(".exists();", "");
        if (start1.equals(end1)) {
            return true;
        }
        System.out.format("THEOREM (%s):\n"+
                          "%s\n=\n%s\n", name, start1, end1);
        System.out.format("details:\n"+ 
                           "       %s =>\n"+
                           "       %s =>\n"+
                           "       %s\n", start1, ss, end1);
        System.out.println(simplify(start1, end1));
        System.out.println();
        return true;
    }

    public String simplLeft, simplRight;

    public String simplify(String start, String end) {
        int i;
        int min = Integer.min(start.length(), end.length());
        for (i = 0; i < min; i++) {
            if (start.charAt(i) != end.charAt(i)) {
                break;
            }
        }
        String st = start.substring(i);
        String en = end.substring(i);
        min = min - i;
        int stmax = st.length();
        int enmax = en.length();
        for (i = 0; i < min; i++) {
            if (st.charAt(stmax - i - 1) != en.charAt(enmax - i - 1)) {
                break;
            }
        }
        st = st.substring(0, stmax - i);
        en = en.substring(0, enmax - i);

        if (st.equals("")) {
            st = "id()";
        }
        if (en.equals("")) {
            en = "id()";
        }

        return String.format("Conjectured theorem: %s  ==  %s", st, en);

//        LinkedList<String> list1 = split(start);
//        LinkedList<String> list2 = split(end);
//        int min = Integer.min(list1.size(), list2.size());
//        for (int i=0; i<min; i++) {
//            if (list1.getFirst().equals(list2.getFirst())) {
//                list1.removeFirst();
//                list2.removeFirst();
//            }
//        }
//        min = Integer.min(list1.size(), list2.size());
//        for (int i=0; i<min; i++) {
//            if (list1.getLast().equals(list2.getLast())) {
//                list1.removeLast();
//                list2.removeLast();
//            }
//        }
//        String x1 = compact(list1);
//        String x2 = compact(list2);
//        
//        return String.format("Conjectured theorem: %s  ==  %s", x1,x2);
    }

    public static LinkedList<String> split(String s) {
        LinkedList<String> list = new LinkedList<>();
        int start = 0;
        int curr = 0;
        int end = s.length();
        int inside = 0;
        while (curr < end) {
            if (s.charAt(curr) == '(') {
                inside++;
            } else if (s.charAt(curr) == ')') {
                inside--;
            } else if ((s.charAt(curr) == '.' || s.charAt(curr) == ';') && inside == 0) {
                String next = s.substring(start, curr);
                list.add(next);
                start = curr + 1;
            }
            curr++;
        }
        return list;
    }

    public String compact(LinkedList<String> ll) {
        if (ll.size() == 0) {
            return "id()";
        }
        String result = "";
        String dot = "";
        for (String l : ll) {
            result = result + dot + l;
            dot = ".";
        }
        return result;
    }
}
