package Common;


import static org.junit.Assert.fail;

public abstract class ParseTst<Sentence extends ISentence> {
    protected Sentence s;
    protected IArrow a;

    abstract public Sentence makeSentence(String s);
    abstract public IArrow makeArrow();

    protected ParseTst() {
        s = makeSentence("");  // just need an instance
        a = makeArrow();
    }

    protected int doitAlles(IArrow arr, String sent) {
        s = makeSentence(sent); // must parse to type Bool
        try {
            s.Parse();
            return 0;
        } catch (RuntimeException e) {
            System.out.println(arr.name() + " FAILED: " + s.origSen() + " : " + e.getMessage());
            return 1;
        }
    }
    
    public void alles() {
        int failures = 0;
        int ntests = 0;
        int etests = 0;
        for (IArrow arr: a.ArrowValues()) {
            String v = arr.getParseTest();
            if (v.startsWith("ERROR")) {
                etests++;
                continue;
            }
            ntests++;
            failures = failures + doitAlles(arr,v + ";");
        }
        System.out.format("%d failures %d non-tests in %d tests\n\n",failures,etests,ntests);
        if (failures != 0) {
            fail();
        }
        // ensures that all expressions parse
    }

    public String doit(String sent) {
        Sentence s = makeSentence(sent); // must parse to a Bool expression
        try {
            s.Parse();
            // if we get here, sent was parsed to be of type Bool
            Boolean consumedAll = s.atEnd();

            if (consumedAll) {
                return s.toString();
            }
            System.out.println("FAILED");
            System.out.println("  input: " + sent);
            System.out.println(" output: " + s.toString());
            System.out.println("remains: " + s.toParse());
            return "not all consumed";
        } catch (RuntimeException e) {

            // if we get here, sent was parsed to be of type Bool
            Boolean consumedAll = s.atEnd();

            if (consumedAll) {
                return s.toString();
            }
            String result = "type WRONG   " + (consumedAll ? "at end" : "NOT at end");
            result = result + "\n " + e.getMessage();
            fail();
            return result;

        }
    }

    public void doitSuccess(String sent) {
        String result = doit(sent);
        if (!result.equals(sent)) {
            System.err.println(result);
            fail();
        }
        System.out.println("Success!  input: " + sent);
        System.out.println("         output: " + result);
    }

    public void doitFail(String sent) {
        try {
            Sentence s = makeSentence(sent); // must be of Domain.Bool type
            s.Parse();
            fail();
        } catch (RuntimeException e) {
            int i = 1;
        }
    }

}
