package XformTools;

import Common.ArrowType;
import MDLUtilities.Utils;
import PrologDB.DB;
import PrologDB.DBSchema;
import PrologDB.Table;
import PrologDB.TableSchema;
import PrologDB.Tuple;
import java.util.LinkedList;

public class Apl2Fpl {

    // these variables are implicit parameters to most static methods of this file.
    static String TOP = null;
    static Table dom, attr, inh, assoc;        // apl tables
    static Table domain, func;                 // fpl tables

    public static LinkedList<String> primDoms;        // list of all primitive domains

    static {
        primDoms = new LinkedList<>();
        primDoms.add("Bool");
        primDoms.add("Int");
        primDoms.add("Str");
        primDoms.add("Var");
        primDoms.add("VarArr");
        primDoms.add("End");
        primDoms.add("Void");
        primDoms.add("Pair");
                
    }

    public static String Void = "Void";     // type for Void
    public static String Str = "Str";       // type for Str
    public static String dualChar = "_";    // special character indicating "dual"
    public static String tableChar = "$";   // special character indicating "table"
    public static String Int = "Int";       // type for Int
    public static String Bool = "Bool";     // type for Bool
    public static String Pair = "Pair";     // type for Pair
    public static String Comp = "Computed"; // type for computed

    // initialize identifier counter;
    static int id = 0;

    static String nextId() {
        return "" + id++;
    }

    // Main Methods 
    public static void top(String... args) {
        TOP = "TOP";
        main(args);
    }

    public static void bot(String... args) {
        TOP = "BOTTOM";
        main(args);
    }
    public static String getAppName(String name) {
        String shorten = name.replace(".apl.pl", "");
        String[] arr = Utils.parseFileName(shorten);
        return arr[0];
    }
    

    static void main(String... args) {

        // Step 1: marquee processing -- note: this hack is because 
        // I believe I stumbled over an error in the Java compiler and can't
        // use marque2In_1out... strange.
        if (!args[0].endsWith(".apl.pl") || !args[1].endsWith(".extra.csv") || !args[2].endsWith(".fpl.pl")) {
            System.out.format("Usage %s <>.apl.pl <>.extra.csv <>.fpl.pl\n", Apl2Fpl.class.getName() );
            System.exit(1);
        }
        String inputFileName = args[0];
        String csvFileName = args[1];
        String outputFileName = args[2];
        String appName = getAppName(args[0]);
        
//        Marquee2In_1Out mark = new Marquee2In_1Out(Apl2Fpl.class, ".apl.pl", ".extra.csv", ".fpl.pl",args);
//        String inputFileName = mark.getInputFileName();
//        String csvFileName = mark.getInputFileName1();
//        String outputFileName = mark.getInputFileName2();
//        String appName = mark.getAppName();

        // Step 2: read database and create empty database.
        DB apl = DB.read(inputFileName);
        dom = apl.getTableEH("domain");
        attr = apl.getTableEH("attr");
        inh = apl.getTableEH("inh");
        assoc = apl.getTableEH("assoc");

        DBSchema as = new DBSchema("f");
        as.add(new TableSchema("domain").add("id", "name", "prim:bool", "computed:bool"));
        as.add(new TableSchema("func").add("id", "name", "input", "output", "arrowtype", 
                                            "argtype", "vartype", "user:bool", "computed:bool","added:bool"));
        as.finishedAndIPropagatedAttributesMyself();

        DB fpl = new DB(appName, as);
        domain = fpl.getTableEH("domain");
        func = fpl.getTableEH("func");
        id = 0;

        // Step 3: create domain table
        initDomainTable();

        // Step 4: add Domain Functions
        addDomainFuncs();

        // Step 5: add attribute functions
        addAttributeFuncs();

        // Step 6: add association functions
        addAssocFuncs();

        // Step 7: add all library functions
        addLibraryFuncs();

        // Step 7: add extra user operations
        addExtraOperations(func, TOP, csvFileName);

        // Step 8: now propagate inheritance
        propInh();

        // Step 9: write out file
//        func.write("func.csv");  //debugging
        fpl.print(outputFileName);
    }

    public static void addAttributeFuncs() {
        for (Tuple d : dom.tuples()) {
            String dname = d.get("name");
            Table datts = attr.select(t -> t.is("domain", dname));
            for (Tuple a : datts.tuples()) {
                String aname = a.get("name");
                String atype = a.get("primtype");
                String comp = a.get("computed");
                ////S name, S input, S output, S arrowType, S argType, S varType, b usr, S computed, S added
                add(aname, dname, atype,           ArrowType.field.name(), "null", "null", true, comp, "FALSE");
                add(aname, tog(dname), tog(atype), ArrowType.noargs.name(),"null", "null", true, comp, "FALSE");
            }
        }
    }

    // initialize domain table by copying from apl to fpl
    public static void initDomainTable() {
        for (Tuple d : dom.tuples()) {
            String id = d.getId();
            String name = d.get("name");
            String prim = d.get("prim").toUpperCase();
            String computed = d.get("computed").toUpperCase();

            if (id.startsWith(tableChar)) {
                domain.add(id, name, prim, computed);
                domain.add(tog(id), tog(name), prim, computed);
            } else {
                domain.add(tog(id), tog(name), prim, computed);
                domain.add(id, name, prim, computed);
            }
        }
    }

    public static String tog(String x) {
        if (x.startsWith(tableChar)) {
            return x.substring(1);
        }
        return tableChar + x;
    }


    /////////////////////////////// func tuple addition methods////////////////////////
    // methods to add tuples to the func table
    // this is for user-defined functions
//    public static void add(String name, String input, String output, String arrowType, String computed, String added) {
//        func.add(nextId(), name, input, output, arrowType, Void, "null", "TRUE", computed,added);
//    }

    // this is for system-defined functions like and and gtr
    public static void add(String name, String input, String output,
            String arrowType, String argType, boolean usr, String computed, String added) {
        func.add(nextId(), name, input, output, arrowType, argType, "null", usr ? "TRUE" : "FALSE", computed,added);
    }

    // this is for system-defined functions like select
    public static void add(String name, String input, String output,
            String arrowType, String argType, String varType, boolean usr, String computed, String added) {
        func.add(nextId(), name, input, output, arrowType, argType, varType, usr ? "TRUE" : "FALSE", computed,added);
    }

    // this adds a boolean-output binary operation
    static void add(String opName, String input, String computed, String added) {
        add(opName, input, Bool, ArrowType.hasArgs.name(), input, false, computed, added);
    }

    // this adds a unary operation
//    static void add(String opName, String input, String output, String computed, String added) {
//        add(opName, input, output, ArrowType.noargs.name(), "null", "null", false, computed, added);
//    }

    /**
     * *****************************************************************
     */
    //
    public static void addDomainFuncs() {
        // Step 4: add void functions for domains
        for (Tuple d : dom.tuples()) {
            String dname = d.get("name"); // should not have $ in front.
            String comp  = d.get("computed");
            //S name, S input, S output, S arrowType, S argType, S varType, b usr, S computed, S added
            add(dname, Void, dname, ArrowType.variable.name(),"null","null", false, "FALSE", "FALSE");
            add(tog(dname), Void, tog(dname), ArrowType.variable.name(), "null","null", false, comp, "FALSE");

            add("exists", tog(dname), Bool, ArrowType.noargs.name(),"null","null",false, "FALSE", "FALSE");
            add("count",  tog(dname), Int, ArrowType.noargs.name(),"null","null",false, "FALSE", "FALSE");
            add("toTable", dname, tog(dname), ArrowType.noargs.name(),"null","null",false, "FALSE", "FALSE");
            add("id", dname, dname, ArrowType.noargs.name(),"null","null",false, "FALSE", "FALSE");
            add("id", tog(dname), tog(dname), ArrowType.noargs.name(),"null","null",false, "FALSE", "FALSE");
            add("select", tog(dname), tog(dname), ArrowType.hasArgs.name(), Bool, dname, false, "FALSE", "FALSE");
        }
    }

    static void addLibraryFuncs() {
        // add library domains
        for (String i : primDoms) {
            if (i.startsWith("V") || i.startsWith("End")) {
                domain.add(i, i, "TRUE", "FALSE");
                ////S name, S input, S output, S arrowType, S argType, S varType, b usr, S computed, S added
                add(i, Void, Void, ArrowType.variable.name(), "null", "null", false, "FALSE", "FALSE");
            } else {
//            add(tableChar+i,Void,tableChar+i,ArrowType.variable.name());
//            add(i,Void,i,ArrowType.variable.name());
            }
        }

        // add standard functions
        add("and", Bool, "FALSE", "FALSE");
        add("eq", Int, "FALSE", "FALSE");
        add("equals", Str, "FALSE", "FALSE");
        add("geq", Int, "FALSE", "FALSE");
        add("leq", Int, "FALSE", "FALSE");
        add("gtr", Int, "FALSE", "FALSE");
        add("lss", Int, "FALSE", "FALSE");
        add("or", Bool, "FALSE", "FALSE");
        //S name, S input, S output, S arrowType, S argType, S varType, b usr, S computed, S added
        add("not", Bool, Bool, ArrowType.hasArgs.name(), Bool, "null", false, "FALSE", "FALSE");
        add("first",Pair,Str,ArrowType.field.name(),"null","null",false,"FALSE","FALSE");
        add("last",Pair,Str,ArrowType.field.name(),"null","null",false,"FALSE","FALSE");
        add("first",tog(Pair),tog(Str),ArrowType.noargs.name(),"null","null",false,"FALSE","FALSE");
        add("last",tog(Pair),tog(Str),ArrowType.noargs.name(),"null","null",false,"FALSE","FALSE");
    }

    // propagate functions from supertables to subtables
    static void propInh() {
        LinkedList<Tuple> roots = roots();
        for (Tuple r : roots) {
            pushDown(r);
        }
    }

    // find roots of inheritance hierarchies
    // these classes have no superclasses
    static LinkedList<Tuple> roots() {
        LinkedList<Tuple> roots = new LinkedList<>();
        for (Tuple d : dom.tuples()) {
            String dname = d.get("name");
            if (inh.select(p -> p.is("name", dname)).isEmpty()) {
                roots.add(d);
            }
        }
        return roots;
    }

    // push down functions in table s to all of its 
    // immediate subtables, and recursive to subtables of subtables
    static void pushDown(Tuple s) {
        String sup = s.get("name");
        for (Tuple c : inh.select(p -> p.is("super", sup)).tuples()) {
            pushDown(s, c);
            // now find children of c, and push c's functions down to them
            Table children = inh.select(p -> p.is("super", c.get("name")));
            for (Tuple cc : children.tuples()) {
                pushDown(c, cc);
            }
        }
    }

    static void pushDown(Tuple s, Tuple c) {
        String sup = s.get("name");
        String chld = c.get("name");
        Table fs = func.select(p -> p.is("input", sup));
        // add each tuple in fs to func with input domain name = chld
        for (Tuple t : fs.tuples()) {
            ////S name, S input, S output, S arrowType, S argType, S varType, b usr, S computed, S added
            add(t.get("name"), chld, t.get("output"), t.get("arrowtype"), t.get("argtype"), t.get("vartype"), t.getBool("user"), t.get("computed"), t.get("added"));
        }
    }

    // for each assoc, create two functions (f+dual)
    public static void addAssocFuncs() {
        // each assoc becomes a function and its dual (for table inputs)
        // each assoc becomes a function and its dual (for tuple inputs)
        for (Tuple l : assoc.tuples()) {
            int id = Integer.parseInt(l.getId());
            //S name, S input, S output, S arrowType, S argType, S varType, b usr, S computed, S added
            add(l.get("name"), l.get("input"), tog(l.get("output")), "noargs", "null", "null", true, l.get("computed"), "FALSE");
            add(l.get("name"), tog(l.get("input")), tog(l.get("output")), "noargs", "null", "null", true, l.get("computed"), "FALSE");
            add(l.get("name") + dualChar, l.get("output"), tog(l.get("input")), "noargs","null", "null", true, l.get("computed"), "FALSE");
            add(l.get("name") + dualChar, tog(l.get("output")), tog(l.get("input")), "noargs", "null", "null", true,l.get("computed"), "FALSE");
        }
    }

//    // this checks to see if name is a table, and returns non-table name 
//    static String tup(String tableName) {
//        if (tableName.startsWith(tableChar)) {
//            return tableName.substring(1);
//        }
//        throw new RuntimeException("table name '" + tableName + "' does not begin with '" + tableChar + "'");
//    }
    static String tab(String tupleName) {
        return tableChar + tupleName;
    }

    // add extra operations
    // top = "T" or "B", although value stored is T,B,or TB */
    public static void addExtraOperations(Table func, String TOP, String XtraCVSName) {
        Table cvs = Table.read(XtraCVSName);
        String T = TOP.equals("TOP")?"T":"B";
        for (Tuple t : cvs.tuples()) {
            String torb = t.get("TorBorTB").toUpperCase();
            if (torb.contains(T)) {
                //S name, S input, S output, S arrowType, S argType, S varType, b usr, S computed, S added
                add(t.get("fname"), t.get("input"), t.get("output"), ArrowType.noargs.name(),"null","null",false,"FALSE", "TRUE");
            }
        }
    }
}
