package XformTools;

import CD.AssocSchema;
import MDLUtilities.Marquee2In_0Out;
import CD.CDSpec;
import PrologDB.Column;
import PrologDB.DBSchema;
import PrologDB.SubTableSchema;
import PrologDB.Table;
import PrologDB.TableSchema;
import PrologDB.Tuple;

public class CDSpec2Apl extends XformCommon {

    static String aplEnd = ".apl.pl";
    static DBSchema aplSchema;
    static CDSpec cd = null;
    static Table csv;

    public static void main(String... args) {
        // Step 1: marquee processing
        Marquee2In_0Out mark = new Marquee2In_0Out(CDSpec2Apl.class, ".cdspec.pl", ".csv", args);
        String cdspecFilename = mark.getInputFileName();
        String csvFilename = mark.getInputFileName1();
//        String appName = mark.getAppName();

        // Step 2: read in arg1.cdspec.pl, and arg2.csv
        cd = null;
        csv = null;
        try {
        cd = CDSpec.read(cdspecFilename);
        } catch (Exception e) {
            System.err.println("unable to parse " + cdspecFilename);
            System.err.println(e.getMessage());
            return;
        } try {
        csv = Table.read(csvFilename);
        } catch (Exception e) {
            System.err.println("unable to parse " + csvFilename);
            System.err.println(e.getMessage());
            return;
        }
        
        main(cd,csv);
    }
    
    public static void main(CDSpec cdspec, Table csvtable) {
        cd = cdspec;
        csv = csvtable;

        // Step 3: initialize
        aplSchema = MakeAplSchema();

        // Step 3: make apl files, top then bottom
        
//        System.out.println("-----------------Top----------------------");
        makeAplFile(true);
//        System.out.println("-----------------Bottom-------------------");
        makeAplFile(false);

//        System.out.println("Done!");
    }

    static void makeAplFile(boolean top) {
        aid = 0;
        String TB = top ? "Top" : "Bottom";
        apl = initAplDB(TB, aplSchema);

        // Step 1: do tables 
        Table domain = apl.getTableEH("domain");
        domain.add("Int", "Int", "TRUE", "FALSE");
        domain.add("Str", "Str", "TRUE", "FALSE");
        domain.add("Bool", "Bool", "TRUE", "FALSE");
        domain.add("Pair", "Pair", "TRUE", "FALSE");
        for (TableSchema ts : cd.getTableSchemas()) {
            String name = ts.getName();
            domain.add(name, name, "FALSE", isComputedClass(top,name));
        }

        // Step 2: do attrs
        Table attr = apl.getTableEH("attr");
        for (TableSchema ts : cd.getTableSchemas()) {
            String cname = ts.getName();
            for (Column c : ts.getColumns()) {
                String name = c.getName();
                String type = c.getKind();
                attr.add(aid(), cname, name, capFirst(type), "field",isComputedAttr(top,cname,name));
            }
        }
        
        // Step 3: do inheritance
        Table inh = apl.getTableEH("inh");
        for (SubTableSchema sts : cd.getSubTableSchemas()) {
            String supr = sts.getSuper().getName();
            for (TableSchema ts : sts.getSubTableSchemas()) {
                String sub = ts.getName();
                inh.add(aid(),supr,sub);
            }
        }
        
        // Step 4: do assoc
        Table assoc = apl.getTableEH("assoc");
        for (AssocSchema as : cd.getAssocSchemas()) {
            String name = as.getName();
            String in   = as.getClass2();
            String out  = as.getClass1();  // this order to match previous implementation
            assoc.add(aid(),name,in,out,isComputedAsso(top,name));
            aid++;  // to match previous implementation
        }
        
        apl.print(TB+aplEnd);
    }
    
    static String aid() {
        return String.format("%03d",aid++);
    }

    static String isComputedAttr(boolean top, String cname, String fld) {
        String TorB = top?"T":"B";
        for (Tuple t : csv.tuples()) {
            int i=0; // for debugging
            if (t.is("CorA","C") && t.is("name", cname) && t.is("field", fld) && t.is("TorB",TorB)) {
                return "TRUE";
            }
        }
        return "FALSE";
    }
    
    static String isComputedClass(boolean top, String className) {
        String TorB = top?"T":"B";
        for (Tuple t : csv.tuples()) {
            int i=0; // for debugging
            if (t.is("CorA","C") && t.is("name",className) && t.is("TorB",TorB)  && t.is("field","")) {
                return "TRUE";
            }
        }
        return "FALSE";
    }
    
        static String isComputedAsso(boolean top, String aName) {
        String TorB = top?"T":"B";
        for (Tuple t : csv.tuples()) {
            int i=0; // for debugging
            if (t.is("CorA","A") && t.is("name", aName) && t.is("TorB",TorB)) {
                return "TRUE";
            }
        }
        return "FALSE";
    }
    // WHY IN THE HELL DID I USE THIS IN THE FIRST PLACE??
    // UPON REVISION REMOVE and REMOVE FROM VPL2APL
    public static String capFirst(String str) {
        if (str == null || str.isEmpty()) {
            return str;
        }

        return str.substring(0, 1).toUpperCase() + str.substring(1);
    }
}
