package XformTools;

import PrologDB.DB;
import PrologDB.Table;
import PrologDB.Tuple;
import java.util.function.BiFunction;


public class UtilIsomorphic {
    static class apl {
        Table domain, attr, inh, assoc;
        
        public apl(DB db) {
            domain = db.getTableEH("domain");
            attr = db.getTableEH("attr");
            inh = db.getTableEH("inh");
            assoc = db.getTableEH("assoc");
        }
    }
    
    static  BiFunction<Tuple,Tuple,Boolean> compDomain = (t1,t2) -> domainTuples(t1,t2);
    static  BiFunction<Tuple,Tuple,Boolean> compAttr   = (t1,t2) -> attrTuples(t1,t2);
    static  BiFunction<Tuple,Tuple,Boolean> compInh    = (t1,t2) -> inhTuples(t1,t2);
    static  BiFunction<Tuple,Tuple,Boolean> compAssoc  = (t1,t2) -> assocTuples(t1,t2);
    
    static void isomorphic(DB left, DB right) {
        apl l = new apl(left);
        apl r = new apl(right);
        iso(l.domain,r.domain, compDomain);
        iso(l.attr, r.attr, compAttr);
        iso(l.inh,r.inh, compInh);
        iso(l.assoc,r.assoc, compAssoc);
    }
            
    static boolean areSame(Tuple left, Tuple right, BiFunction<Tuple,Tuple,Boolean> f) {
        return f.apply(left,right);
    }  
    
    static void iso(Table left, Table right, BiFunction<Tuple,Tuple,Boolean> f) {
        if (left.size()!=right.size()) {
            throw new RuntimeException(String.format("%s tables have different sizes", left.getName()));
        }
        for (Tuple l : left.tuples()) {
            boolean match = false;
            for (Tuple r: right.tuples()) {
                if (areSame(l,r,f)) {
                    match = true;
                    break;
                }
            }
            if (!match)
                throw new RuntimeException(
                        String.format("%s tables do not have identical tuples with id=%s",left.getName(),l.getId()));
        }
    }
    
    static boolean domainTuples(Tuple l, Tuple r) {
        return comp(l,r,"name") && comp(l,r,"prim");
    }
    
    static boolean attrTuples(Tuple l, Tuple r) {
        boolean r1 = comp(l,r,"domain") && comp(l,r,"name");
//        boolean r2 = l.is("primtype","Computed") || r.is("primtype","Computed");
//        if (!r2) {
        boolean r2 = comp(l,r,"primtype");
//        }
        return r1 && r2;
    }
    
    static boolean inhTuples(Tuple l, Tuple r) {
        return comp(l,r,"name") && comp(l,r,"super");
    }
    
    static boolean assocTuples(Tuple l, Tuple r) {
        return comp(l,r,"name") && comp(l,r,"input") && comp(l,r,"output");
    }
    
    static boolean comp(Tuple l, Tuple r, String field) {
        return l.is(field,r.get(field));
    }
}
