package XformTools;

import Common.ArrowType;
import MDLUtilities.Marquee1In_1Out;
import PrologDB.DB;
import PrologDB.DBSchema;
import PrologDB.ErrorReport;
import PrologDB.Table;
import PrologDB.TableSchema;
import PrologDB.Tuple;
import static XformTools.Apl2Fpl.tableChar;
import java.util.HashSet;


public class Vpl2Apl extends XformCommon {
    
    public static final String computedChar = "C";

    static String[] prims = {"int", "str", "bool", "pair"};

    static boolean isPrim(String v, String f, String t) {
        for (String p : prims) {
            if (t.equals(p)) {
                return true;
            }
        }
        return false;
    }

    static boolean isPrim(String v, String f, String t, ErrorReport er) {
        boolean result = isPrim(v, f, t);
        if (!result) {
            er.add("unrecognized attribute type %s.%s:%s", v, f, t);
        }
        return result;
    }

    public static String capFirst(String str) {
        if (str == null || str.isEmpty()) {
            return str;
        }

        return str.substring(0, 1).toUpperCase() + str.substring(1);
    }
    
    static String getAid() {
        return String.format("%03d",aid++);
    }
    
//    protected static ErrorReport er;
//    protected static DB apl;
//    protected static Table domain, attr, inh, assoc;
//    protected static int aid = 0;

//    static String getAid() {
//        return "" + aid++;
//    }

    public static void main(String... args) {

        // Step 1: marquee processing
        Marquee1In_1Out mark = new Marquee1In_1Out(Vpl2Apl.class, ".vpl.pl", ".apl.pl", args);
        String inputFileName = mark.getInputFileName();
        String outputFileName = mark.getOutputFileName();
        String appName = mark.getAppName();
        er = new ErrorReport();

        // Step 2: read database and create empty database.
        DB vpl = DB.read(inputFileName);
        Table vbox = vpl.getTableEH("vBox");
        Table vassoc = vpl.getTableEH("vAssociation");

        DBSchema as = MakeAplSchema();
        apl = initAplDB(appName, as);

        // Step 3: create domain table
        HashSet<String> uniqueName = new HashSet<>();
        for (String p : prims) {
            String computed = "FALSE";
            p = capFirst(p);
            domain.add(p, p,"TRUE",computed);
            uniqueName.add(p);
        }
        for (Tuple v : vbox.tuples()) {
            String id = v.get("name");
            if (id.startsWith(tableChar)) {
                er.add("class name %s should not start with %c",id,tableChar);
            }
            if (!id.equals(capFirst(id))) {
                er.add("class name not capitalized: %s", id);
            }
            if (uniqueName.contains(id)) {
                er.add("duplicate class name %s", id);
                continue;
            }
            String computed = !v.is("methods","")?"TRUE":"FALSE";
            domain.add(id, id, "FALSE",computed);
            uniqueName.add(id);
        }

        // Step 4: create attributes
        for (Tuple v : vbox.tuples()) {
            uniqueName = new HashSet<>();
            String vname = v.get("name");
            String fields = v.get("fields");
            String f[] = fields.split("%");
            for (String s : f) {
                String nt[] = s.split(":");
                if (nt.length>2) {
                    er.add("illegal field definition ("+s+")");
                    continue;
                }
                String fname = nt[0];
                String type = (nt.length == 1) ? "str" : nt[1];
                String computed = "FALSE";
                if (type.startsWith(computedChar)) {
                    computed = "TRUE";
                    type = type.substring(1);
                }
                boolean ip = isPrim(vname, fname, type, er);
                type = capFirst(type);
                if (uniqueName.contains(fname)) {
                    er.add("duplicate field %s in table %s", fname, vname);
                    continue;
                } else {
                    uniqueName.add(fname);
                    attr.add(getAid(), vname, fname, type, ArrowType.field.name(),computed);
                }
            }
        }

        // Step 5: create inheritance
        for (Tuple t : vassoc.tuples()) {
            // if this is an inheritance, ensure no roles
            if (!t.is("arrow1", "TRIANGLE") && !t.is("arrow2", "TRIANGLE")) {
                continue;
            }
            boolean errors = false;
            if (!t.is("role1", "") || !t.is("role2", "")) {
                er.add("vAssociation id=%s is inheritance with roles", t.getId());
                errors = true;
            }
            if (!t.is("middleLabel","")) {
                er.add("vAssociation %s is inheritance -- no middleLabel", t.getId());
                errors = true;
            }
            if (!t.is("lineStyle","")) {
                er.add("vAssociation %s is inheritance -- no DOTTED lineStyle", t.getId());
                errors = true;
            }
            if (t.is("arrow1", "TRIANGLE") && t.is("arrow2", "TRIANGLE")) {
                er.add("Bidirectional inheritance Vassociation.id=%s disallowed", t.getId());
                errors = true;
            }
            if (errors)
                continue;

            if (t.is("arrow1", "TRIANGLE")) {
                String supname = vbox.follow(t.get("cid1")).get("name");
                String name = vbox.follow(t.get("cid2")).get("name");
                inh.add(getAid(), name, supname);
                continue;
            }
            if (t.is("arrow2", "TRIANGLE")) {
                String supname = vbox.follow(t.get("cid2")).get("name");
                String name = vbox.follow(t.get("cid1")).get("name");
                inh.add(getAid(), name, supname);
                continue;
            }
        }

        // Step 6: create associations
        uniqueName = new HashSet<>();
        for (Tuple t : vassoc.tuples()) {
            if (t.is("arrow1", "TRIANGLE") || t.is("arrow2", "TRIANGLE")) {
                continue;
            }
            String name = t.get("middleLabel");
            if (name.equals("")) {
                er.add("vAssociation id=%s has empty middle label", t.getId());
                continue;
            }
            if (uniqueName.contains(name)) {
                er.add("multiple vAssociations have the same name=%s", name);
                continue;
            }
            uniqueName.add(name);
            String computed = t.is("lineStyle", "DOTTED") ? "TRUE" : "FALSE";
//            if (!t.is("role1", "") && !t.is("role2", "")) {
//                er.add("role names present in vAssociation %s", name);
//                continue;
//            }
            if (t.is("arrow1", "V") && t.is("arrow2", "V")) {
                er.add("vAssociation %s goes in <-> both directions", name);
                continue;
            }
            String name1 = vbox.follow(t.get("cid1")).get("name");
            String name2 = vbox.follow(t.get("cid2")).get("name");
            if (t.is("arrow1", "V")) {
                assoc.add(getAid(), name, name2, name1, computed);
                getAid(); /// skip counter to add dual later
                continue;
            }
            if (t.is("arrow2", "V")) {
                assoc.add(getAid(), name, name1, name2, computed);
                getAid(); /// skip counter to add dual later
                continue;
            }
            er.add("vAssociation %s has no arrow/direction", name);
        }
        er.printEH();
        apl.print(outputFileName);
//        System.out.format("%s written\n", outputFileName);
    }

//    static DB initAplDB(String appName, DBSchema as) {
//        apl = new DB(appName, as);
//        domain = apl.getTableEH("domain");
//        attr = apl.getTableEH("attr");
//        inh = apl.getTableEH("inh");
//        assoc = apl.getTableEH("assoc");
//        aid = 0;
//        return apl;
//    }
//
//    static DBSchema MakeAplSchema() {
//        DBSchema as = new DBSchema("all");
//        as.add(new TableSchema("domain").add("id", "name", "prim:bool","computed:bool"));
//        as.add(new TableSchema("attr").add("id", "domain", "name", "primtype","arrowtype","computed:bool"));
//        as.add(new TableSchema("inh").add("id", "name", "super"));
//        as.add(new TableSchema("assoc").add("id", "name", "input", "output", "computed:bool"));
//        as.finishedAndIPropagatedAttributesMyself();
//        return as;
//    }

}
