package jrefact;

import Common.ArrowType;
import MDLUtilities.RunningBear;
import PrologDB.Table;
import PrologDB.Tuple;
import static XformTools.Apl2Fpl.tableChar;
import static XformTools.Apl2Fpl.primDoms;
import java.util.Collections;
import java.util.LinkedList;

import static jrefact.JRefact.func;
import static jrefact.JRefact.domain;

class JArrow extends RunningBear {

    static class tri {

        public final String name, in, out;

        public tri(String n, String i, String o) {
            name = n;
            in = i;
            out = o;
        }
    }

    static final String unknownFunction = "ERROR! unknown operation -- shouldn't occur";
    static final String Error = "\"ERROR!\"";
    static final String indent = "\n                                         ";
    static final String SUPPLY = "\"XXX!\"";

    static int id;
    static int ln;
    static Table extra;
    static LinkedList<tri> functorsToComplete, dualFunctorsToComplete;
    static boolean botm; // for debugging only

    public static void main(String dir, boolean bottom, Table xcsv, Table ucsv, Table diffcsv) {
        botm = bottom;    // for debugging only
        extra = xcsv;
        id = 10;
        ln = 0;
        openOut(dir + "/Arrow.java");
        header(bottom);
        genArrows();
        footer(bottom);
        closeOut();
        genPatch(JRefact.CommonSrc, bottom, xcsv, ucsv, diffcsv);
    }

    static void genPatch(String dir, boolean bot, Table xcsv, Table ucsv, Table diffcsv) {
        String xxx;
        bot = !bot;
        String b = bot ? "Bottom" : "Top";
        openOut(dir + "/" + b + "Update.java");
        l("package CEXCommon;");
        l();
        l("public class %sUpdate {", b);
        l("   public static void update() {");
        // stolen from https://dzone.com/articles/using-lambda-expression-sort
        Collections.sort(functorsToComplete, (tri p1, tri p2) -> p1.name.compareTo(p2.name));
        for (tri s : functorsToComplete) {
            xxx = searchUcsv(ucsv,b,s.name,s.in,s.out,diffcsv);
            l("      %s.Arrow.find(%12s,%s.Domain.%s,%s.Domain.%s).set(\"%s\");", b, "\"" + s.name + "\"", b, s.in, b, s.out, xxx);
        }
        l("   }");
        l(2);
        Collections.sort(dualFunctorsToComplete, (tri p1, tri p2) -> p1.name.compareTo(p2.name));
        b = bot ? "Top" : "Bottom";
        l("   public static void dualUpdate() {");
        for (tri s : dualFunctorsToComplete) {
            xxx = searchUcsv(ucsv,b,s.name,s.in,s.out,diffcsv);
            l("      %s.Arrow.find(%12s,%s.Domain.%s,%s.Domain.%s).set(\"%s\");", b, "\"" + s.name + "\"", b, s.in, b, s.out,xxx);
        }
        l("   }");
        l("}");
        closeOut();
    }
    
    // references warningCount in BETA.MainOld
    static String searchUcsv(Table ucsv, String b, String f, String i, String o, Table diffcsv) {
        String TorB = ""+ b.charAt(0);
        Tuple t = ucsv.getFirst(r->r.is("TorB",TorB) && r.is("fname",f) && r.is("input",i) && r.is("output",o) );
        if (t==null) {
            if (AFunctor.MainOld.warningCount++==0) {
                System.out.println();
                System.out.format("Warning: add and complete these tuples to the .update.csv file:\n");
                System.out.format("TorB,fname,input,output,set,Notes\n");
                System.out.format("---------------------------------\n");
            }
            System.out.format("%s,%s,%s,%s,xxx,\n", TorB,f,i,o);
            return "xxx";
        }
        return t.get("set");
    }

    static boolean isPrimType(String x) {
        for (String s : primDoms) {
            if (s.equals(x)) {
                return true;
            }
        }
        return false;
    }

    static void addTo(Tuple a, String n, String i, String o) {
        functorsToComplete.add(new tri(n, i, o));
    }

    static void addToDual(Tuple a, String n, String i, String o) {
        dualFunctorsToComplete.add(new tri(n, i, o));
    }

    static void genArrows() {
        functorsToComplete = new LinkedList<>();
        dualFunctorsToComplete = new LinkedList<>();
        String ename = null;
        int enumCounter = 1;
        Table sfunc = func.select(f -> f.is("input", "Void"));
//        sfunc.print();  // debugging
//        sfunc.select(f->f.is("output","Void")).print();  // debugging

        // iterate over each $table and tuple type
        // print header then methods
        for (Tuple t : sfunc.tuples()) {
            String output = t.get("output");
            if (output.equals("Void")) {
                String name = t.get("name");
                if (name.equals("Void")) continue;
                l();
                l("    //  Tuple %s", name);
                l("    %-15s(Void, Void, variable, %d, false, \"%s\",",name+"000",enumCounter++,name);
                l("                                         \"ERROR! unknown operation -- shouldn't occur\"),");
                continue;
            }

            String what = (output.startsWith(tableChar)) ? "Table" : "Tuple";
            l();
            l("    //  %s %s", what, output);

            if (isPrimType(output)) {
                ename = String.format("%s000", output);
            } else {
                ename = String.format("%s%03d", output, enumCounter);
            }
            boolean fcomputed = t.getBool("computed");
            boolean tcomputed = domain.follow(output).getBool("computed");
            boolean computed = fcomputed || tcomputed;
            if (computed) {
                if (output.startsWith("$")) {
                    addTo(t, output, t.get("input"), t.get("output"));
                }
                l("    %-15s(%s, %s, %s, %d, %s, %s,%s%s),",
                        ename, t.get("input"), t.get("output"), t.get("arrowtype"),
                        enumCounter++, t.get("user").toLowerCase(),
                        functorEntry(t), indent, Error);
            } else {
                l("    %-15s(%s, %s, %s, %d, %s, %s,%s%s),",
                        ename, t.get("input"), t.get("output"), t.get("arrowtype"),
                        enumCounter++, t.get("user").toLowerCase(),
                        functorEntry(t), indent, parseEntry(t));
            }
            // methods of $table or tuple type
            for (Tuple f : func.select(f -> f.is("input", output)).tuples()) {
                String fname = f.get("name");

                // rule: if a field is computed, don't enter it in an enum
                if (f.getBool("computed") && f.is("arrowtype", "field")) {
                    addTo(f, fname, f.get("input"), f.get("output"));
                    continue;
                }

//                if (fname.startsWith("semi")) {
//                    int i = 1;
//                }
                if (f.getBool("computed")) {
                    addTo(f, fname, f.get("input"), f.get("output"));
                    continue;  // do not add operations that are computed
                }                    // why?  they don't exist in the base language of the model  and were added
                if (f.getBool("added")) {
                    addToDual(f, fname, f.get("input"), f.get("output"));  // add extra operations
                }
                ename = String.format("%s%03d", fname, enumCounter);
                switch (f.get("arrowtype")) {
                    case "variable":
                    case "noargs":
                    case "field":
                        l("    %-15s(%s, %s, %s, %d, %s, %s,%s%s),",
                                ename, f.get("input"), f.get("output"), f.get("arrowtype"),
                                enumCounter++, f.get("user").toLowerCase(),
                                functorEntry(f), indent, computed ? Error : parseEntry(f));
//                        addTo(f, fname, f.get("input"), f.get("output"));
                        break;
                    case "hasArgs":
                        if (!fname.equals("select")) {
                            l("    %-15s(%s, %s, %s, %s, %d, %s, %s,%s%s),",
                                    ename, f.get("input"), f.get("output"), f.get("arrowtype"),
                                    f.get("argtype"),
                                    enumCounter++, f.get("user").toLowerCase(),
                                    functorEntry(f), indent, computed ? Error : parseEntry(f));
//                            addTo(f, fname, f.get("input"), f.get("output"));
                        } else {
                            l("    %-15s(%s, %s, %s, %s, %s, %d, %s, %s,%s%s),",
                                    ename, f.get("input"), f.get("output"), f.get("arrowtype"),
                                    f.get("argtype"), f.get("vartype"),
                                    enumCounter++, f.get("user").toLowerCase(),
                                    functorEntry(f), indent, computed ? Error : parseEntry(f));
//                            addTo(f, fname, f.get("input"), f.get("output"));
                        }
                        break;
                    default:
                        throw new RuntimeException("unknown arrowtype = " + f.get("arrowtype"));
                }
            }
        }
        l("    ;");
        l();
    }

    static void header(boolean bottom) {
        String pkname = bottom ? "Bottom" : "Top";
        l("package %s;", pkname);
        l();
        l("import static %s.Domain.*;", pkname);
        l("import Common.ArrowType;");
        l("import static Common.ArrowType.*;");
        l("import Common.ParseArrowType;");
        l("import Common.IArrow;");
        l(2);
        l("public enum Arrow implements Common.IArrow<Arrow,Domain> {");
    }

    static void footer(boolean bot) {
        String b = bot ? "Bottom" : "Top";
        Append(Common.Footers.ArrowFooter);
        l("   static { CEXCommon.%sUpdate.update();", b);
        b = bot ? "Top" : "Bottom";
        l("            CEXCommon.%sUpdate.dualUpdate();", b);
        l("   }");
        l();
        l("}");
    }

    private static String parseEntry(Tuple a) {
        return "\"" + basicParseEntry(a) + "\"";
    }

    private static String basicParseEntry(Tuple a) {
        ////////////////////////////////////  VARIABLE //////////////////////////////
        if (a.get("arrowtype").equals(ArrowType.variable.name())) {
            if (a.get("input").equals("Void")) {
                if (a.get("output").startsWith(tableChar)) {
                    return String.format("%s.exists()", a.get("output"));
                } else {
                    String o = a.get("output");
                    if (o.equals("Void")) {
                        return unknownFunction;
                    } else {
                        return String.format("$%s.select(t->t.toTable().exists()).exists()", a.get("output"));
                    }
                }
            }
            if (a.get("input").startsWith(tableChar)) {
                switch (a.get("output")) {
                    case "Str":
                        return String.format("%s.select(t->t.%s.equals(\\\"d\\\")).exists()", a.get("input"), a.get("name"));
                    case "Int":
                        return String.format("%s.select(t->t.%s.eq(2)).exists()", a.get("input"), a.get("name"));
                    case "Bool":
                        return String.format("%s.select(t->t.%s).exists()", a.get("input"), a.get("name"));
                    default:
                        throw new RuntimeException("unknown output " + a.get("output"));
                }
            } else {
                switch (a.get("output")) {
                    case "Str":
                        return String.format("$%s.select(t->t.%s.equals(\\\"d\\\")).exists()", a.get("input"), a.get("name"));
                    case "Int":
                        return String.format("$%s.select(t->t.%s.eq(2)).exists()", a.get("input"), a.get("name"));
                    case "Bool":
                        return String.format("$%s.select(t->t.%s).exists()", a.get("input"), a.get("name"));
                    default:
                        throw new RuntimeException("unknown output " + a.get("output"));
                }
            }
        }

        ////////////////////////////////////  NOARGS //////////////////////////////
        if (a.get("arrowtype").equals(ArrowType.noargs.name())) {
            if (a.get("input").startsWith(tableChar)) {
                if (a.get("output").startsWith(tableChar)) {
                    return String.format("%s.%s().exists()", a.get("input"), a.get("name"));
                } else {
                    switch (a.get("output")) {
                        case "Str":
                            return String.format("%s.%s().equals(\\\"d\\\")", a.get("input"), a.get("name"));
                        case "Int":
                            return String.format("%s.%s().eq(5)", a.get("input"), a.get("name"));
                        case "Bool":
                            return String.format("%s.%s()", a.get("input"), a.get("name"));
                        default:
                            return String.format("%s.%s().exists()", a.get("input"), a.get("name"));
                    }
                }
            } else {
                if (a.get("output").startsWith(tableChar)) {
                    return String.format("$%s.select(t->t.%s().exists()).exists()", a.get("input"), a.get("name"));
                } else {
                    return String.format("$%s.select(t->t.%s().toTable().exists()).exists()", a.get("input"), a.get("name"));
                }
            }
        }

        ////////////////////////////////////  HASARGS //////////////////////////////
        if (a.get("arrowtype").equals(ArrowType.hasArgs.name())) {
            if (a.get("input").startsWith(tableChar)) {
                if (a.get("output").startsWith(tableChar)) {
                    return String.format("%s.%s(t->t.toTable().exists()).exists()", a.get("input"), a.get("name"));
                    //return unknownFunction;
                } else {
//                    if (a.get("name").equals("select")) {
//                        return String.format("%s.select(t->t.toTable().exists()).exists()", a.get("input"), a.get("name"));
//                    }
                    switch (a.get("output")) {
                        case "Str":
                            return String.format("%s.%s(\\\"d\\\").exists()", a.get("input"), a.get("name"));
                        case "Int":
                            return String.format("%s.%s(\"5\")", a.get("input"), a.get("name"));
                        case "Bool":
                            return String.format("%s.not()", a.get("input"), a.get("name"));
                        default:
                            return unknownFunction;
                    }
                }
            }
            if (a.get("output").startsWith(tableChar)) {
                return unknownFunction;
            } else {
                if (a.get("output").equals("Bool")) {
                    switch (a.get("input")) {
                        case "Str":
                            return String.format("\\\"d\\\".equals(\\\"d\\\")");
                        case "Int":
                            return String.format("5.%s(5)", a.get("name"));
                        case "Bool":
                            return String.format("5.eq(5).%s(4.eq(4))", a.get("name"));
                        default:
                            return unknownFunction;
                    }
                }
            }
        }

        ////////////////////////////////////  Field //////////////////////////////
        if (a.get("arrowtype").equals(ArrowType.field.name())) {
            switch (a.get("output")) {
                case "Str":
                    return String.format("$%s.select(t->t.%s.equals(\\\"d\\\")).exists()", a.get("input"), a.get("name"));
                case "Int":
                    return String.format("$%s.select(t->t.%s.eq(5)).exists()", a.get("input"), a.get("name"));
                case "Bool":
                    return String.format("$%s.select(t->t.%s).exists()", a.get("input"), a.get("name"));
                case "Pair":
                    return String.format("$%s.select(t->t.%s.first.equals(\\\"a\\\")).exists()", a.get("input"), a.get("name"));

                default:
                    throw new RuntimeException("not primitive type: " + a.get("output"));
            }
        }

        ////////////////////////////////////  End //////////////////////////////  
        return unknownFunction;
    }

    static final char q = '"';

    static String functorEntry(Tuple a) {
//        if (a.getBool("computed")) {
//            return SUPPLY;
//        }
        switch (ArrowType.convert(a.get("arrowtype"))) {
            case variable:
            case field:
                return q + a.get("name") + q;
            case noargs:
                return q + a.get("name") + "()" + q;
            case hasArgs:
                return q + a.get("name") + "(" + q;
            default:
                throw new RuntimeException("unrecognizable arrowtype '" + a.get("arrowtype") + "'");
        }
    }
}
